"""
brozzler/pywb.py - pywb customizations for brozzler including rethinkdb index,
loading from warcs still being written to, canonicalization rules matching
brozzler conventions, support for screenshot: and thumbnail: urls

Copyright (C) 2016-2017 Internet Archive

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import argparse
import json
import sys

import doublethink
import rethinkdb as rdb
import structlog
import urlcanon

import brozzler

logger = structlog.get_logger(logger_name=__name__)

try:
    import pywb.apps.cli
    import pywb.cdx.cdxdomainspecific
    import pywb.cdx.cdxobject
    import pywb.cdx.cdxserver
    import pywb.framework.basehandlers
    import pywb.rewrite.wburl
    import pywb.webapp.query_handler
except ImportError as e:
    logger.critical(
        '%s: %s\n\nYou might need to run "pip install '
        'brozzler[easy]".\nSee README.rst for more information.',
        type(e).__name__,
        e,
    )
    sys.exit(1)


r = rdb.RethinkDB()


class RethinkCDXSource(pywb.cdx.cdxsource.CDXSource):
    def __init__(self, servers, db, table):
        self.servers = servers
        self.db = db
        self.table = table

    @property
    def rr(self):
        try:
            return self._rr
        except AttributeError:
            self._rr = doublethink.Rethinker(self.servers, self.db)
            return self._rr

    def load_cdx(self, cdx_query):
        # logging.debug('vars(cdx_query)=%s', vars(cdx_query))
        rethink_results = self._query_rethinkdb(cdx_query)
        return self._gen_cdx_lines(rethink_results)

    def _gen_cdx_lines(self, rethink_results):
        for record in rethink_results:
            # XXX inefficient, it gets parsed later, figure out how to
            # short-circuit this step and create the CDXObject directly
            blob = {
                "url": record["url"],
                "status": str(record["response_code"]),
                "digest": record["sha1base32"],
                "length": str(record.get("record_length", "-")),
                "offset": str(record["offset"]),
                "filename": record["filename"],
            }
            if record["warc_type"] != "revisit":
                blob["mime"] = record["content_type"] or "-"
            else:
                blob["mime"] = "warc/revisit"
            # b'org,archive)/ 20160427215530 {"url": "https://archive.org/", "mime": "text/html", "status": "200", "digest": "VILUFXZD232SLUA6XROZQIMEVUPW6EIE", "length": "16001", "offset": "90144", "filename": "ARCHIVEIT-261-ONE_TIME-JOB209607-20160427215508135-00000.warc.gz"}'
            cdx_line = "{} {:%Y%m%d%H%M%S} {}".format(
                record["canon_surt"], record["timestamp"], json.dumps(blob)
            )
            yield cdx_line.encode("utf-8")

    def _query_rethinkdb(self, cdx_query):
        start_key = cdx_query.key.decode("utf-8")
        end_key = cdx_query.end_key.decode("utf-8")
        reql = self.rr.table(self.table).between(
            [start_key[:150], r.minval],
            [end_key[:150], r.maxval],
            index="abbr_canon_surt_timestamp",
            right_bound="closed",
        )
        reql = reql.order_by(index="abbr_canon_surt_timestamp")
        # TODO support for POST, etc
        # http_method='WARCPROX_WRITE_RECORD' for screenshots, thumbnails
        reql = reql.filter(
            lambda capture: r.expr(["WARCPROX_WRITE_RECORD", "GET"]).contains(
                capture["http_method"]
            )
        )
        reql = reql.filter(
            lambda capture: (
                (capture["canon_surt"] >= start_key) & (capture["canon_surt"] < end_key)
            )
        )
        if cdx_query.limit:
            reql = reql.limit(cdx_query.limit)
        logger.debug("rethinkdb query", query=reql)
        results = reql.run()
        return results


class TheGoodUrlCanonicalizer(object):
    """
    Replacement for pywb.utils.canonicalize.UrlCanonicalizer that produces
    surts with scheme and with trailing comma, and does not "massage"
    www.foo.org into foo.org.
    """

    def __init__(self, surt_ordered=True):
        """We are always surt ordered (surt_ordered param is ignored)"""
        self.surt_ordered = True

    def __call__(self, url):
        try:
            key = urlcanon.semantic(url).surt().decode("ascii")
            # logging.debug('%s -> %s', url, key)
            return key
        except Exception:
            return url

    def replace_default_canonicalizer():
        """Replace parent class of CustomUrlCanonicalizer with this class."""
        pywb.cdx.cdxdomainspecific.CustomUrlCanonicalizer.__bases__ = (
            TheGoodUrlCanonicalizer,
        )

    def good_surts_from_default(default_surt):
        """
        Takes a standard surt without scheme and without trailing comma, and
        returns a list of "good" surts that together match the same set of
        urls. For example:

             good_surts_from_default('com,example)/path')

        returns

            ['http://(com,example,)/path',
             'https://(com,example,)/path',
             'http://(com,example,www,)/path',
             'https://(com,example,www,)/path']

        """
        if default_surt == "":
            return [""]

        parts = default_surt.split(")", 1)
        if len(parts) == 2:
            orig_host_part, path_part = parts
            good_surts = [
                "http://(%s,)%s" % (orig_host_part, path_part),
                "https://(%s,)%s" % (orig_host_part, path_part),
                "http://(%s,www,)%s" % (orig_host_part, path_part),
                "https://(%s,www,)%s" % (orig_host_part, path_part),
            ]
        else:  # no path part
            host_part = parts[0]
            good_surts = [
                "http://(%s" % host_part,
                "https://(%s" % host_part,
            ]
        return good_surts

    def monkey_patch_dsrules_init():
        orig_init = pywb.cdx.cdxdomainspecific.CDXDomainSpecificRule.__init__

        def cdx_dsrule_init(self, url_prefix, rules):
            good_surts = []
            url_prefixes = [url_prefix] if isinstance(url_prefix, str) else url_prefix
            for bad_surt in url_prefixes:
                good_surts.extend(
                    TheGoodUrlCanonicalizer.good_surts_from_default(bad_surt)
                )
            if "match" in rules and "regex" in rules["match"]:
                rules["match"]["regex"] = r"https?://\(" + rules["match"]["regex"]
            orig_init(self, good_surts, rules)

        pywb.cdx.cdxdomainspecific.CDXDomainSpecificRule.__init__ = cdx_dsrule_init


def support_in_progress_warcs():
    """
    Monkey-patch pywb.warc.pathresolvers.PrefixResolver to include warcs still
    being written to (warcs having ".open" suffix). This way if a cdx entry
    references foo.warc.gz, pywb will try both foo.warc.gz and
    foo.warc.gz.open.
    """
    _orig_prefix_resolver_call = pywb.warc.pathresolvers.PrefixResolver.__call__

    def _prefix_resolver_call(self, filename, cdx=None):
        raw_results = _orig_prefix_resolver_call(self, filename, cdx)
        results = []
        for warc_path in raw_results:
            results.append(warc_path)
            results.append("%s.open" % warc_path)
        return results

    pywb.warc.pathresolvers.PrefixResolver.__call__ = _prefix_resolver_call


class SomeWbUrl(pywb.rewrite.wburl.WbUrl):
    def __init__(self, orig_url):
        import six
        from pywb.rewrite.wburl import WbUrl
        from six.moves.urllib.parse import quote

        pywb.rewrite.wburl.BaseWbUrl.__init__(self)

        if six.PY2 and isinstance(orig_url, six.text_type):
            orig_url = orig_url.encode("utf-8")
            orig_url = quote(orig_url)

        self._original_url = orig_url

        if not self._init_query(orig_url):
            if not self._init_replay(orig_url):
                raise Exception("Invalid WbUrl: ", orig_url)

        new_uri = WbUrl.to_uri(self.url)

        self._do_percent_encode = True

        self.url = new_uri

        # begin brozzler changes
        if (
            self.url.startswith("urn:")
            or self.url.startswith("screenshot:")
            or self.url.startswith("thumbnail:")
        ):
            return
        # end brozzler changes

        # protocol agnostic url -> http://
        # no protocol -> http://
        # inx = self.url.find('://')
        inx = -1
        m = self.SCHEME_RX.match(self.url)
        if m:
            inx = m.span(1)[0]

        # if inx < 0:
        # check for other partially encoded variants
        #    m = self.PARTIAL_ENC_RX.match(self.url)
        #    if m:
        #        len_ = len(m.group(0))
        #        self.url = (urllib.unquote_plus(self.url[:len_]) +
        #                    self.url[len_:])
        #        inx = self.url.find(':/')

        if inx < 0:
            self.url = self.DEFAULT_SCHEME + self.url
        else:
            inx += 2
            if inx < len(self.url) and self.url[inx] != "/":
                self.url = self.url[:inx] + "/" + self.url[inx:]


def _get_wburl_type(self):
    return SomeWbUrl


def monkey_patch_wburl():
    pywb.framework.basehandlers.WbUrlHandler.get_wburl_type = _get_wburl_type


class BrozzlerWaybackCli(pywb.apps.cli.WaybackCli):
    def _extend_parser(self, arg_parser):
        super()._extend_parser(arg_parser)
        arg_parser._actions[4].help = argparse.SUPPRESS  # --autoindex
        arg_parser.formatter_class = argparse.RawDescriptionHelpFormatter
        arg_parser.epilog = """
Run pywb like so:

    $ PYWB_CONFIG_FILE=pywb.yml brozzler-wayback

See README.rst for more information.
"""


# copied and pasted from cdxdomainspecific.py, only changes are commented as
# such below
def _fuzzy_query_call(self, query):
    # imports added here for brozzler
    from pywb.utils.loaders import to_native_str
    from six.moves.urllib.parse import urlsplit, urlunsplit

    matched_rule = None

    urlkey = to_native_str(query.key, "utf-8")
    url = query.url
    filter_ = query.filters

    for rule in self.rules.iter_matching(urlkey):
        m = rule.regex.search(urlkey)
        if not m:
            continue

        matched_rule = rule

        groups = m.groups()
        for g in groups:
            for f in matched_rule.filter:
                filter_.append(f.format(g))

        break

    if not matched_rule:
        return None

    repl = "?"
    if matched_rule.replace:
        repl = matched_rule.replace

    inx = url.find(repl)
    if inx > 0:
        url = url[: inx + len(repl)]

    # begin brozzler changes
    if matched_rule.match_type == "domain":
        orig_split_url = urlsplit(url)
        # remove the subdomain, path, query and fragment
        host = orig_split_url.netloc.split(".", 1)[1]
        new_split_url = (orig_split_url.scheme, host, "", "", "")
        url = urlunsplit(new_split_url)
    # end brozzler changes

    params = query.params
    params.update({"url": url, "matchType": matched_rule.match_type, "filter": filter_})

    if "reverse" in params:
        del params["reverse"]

    if "closest" in params:
        del params["closest"]

    if "end_key" in params:
        del params["end_key"]

    return params


def monkey_patch_fuzzy_query():
    pywb.cdx.cdxdomainspecific.FuzzyQuery.__call__ = _fuzzy_query_call


# copied and pasted from pywb/utils/canonicalize.py, only changes are commented
# as such
def _calc_search_range(url, match_type, surt_ordered=True, url_canon=None):
    # imports added here for brozzler
    import six.moves.urllib.parse as urlparse
    from pywb.utils.canonicalize import UrlCanonicalizeException, UrlCanonicalizer

    def inc_last_char(x):
        return x[0:-1] + chr(ord(x[-1]) + 1)

    if not url_canon:
        # make new canon
        url_canon = UrlCanonicalizer(surt_ordered)
    else:
        # ensure surt order matches url_canon
        surt_ordered = url_canon.surt_ordered

    start_key = url_canon(url)

    if match_type == "exact":
        end_key = start_key + "!"

    elif match_type == "prefix":
        # add trailing slash if url has it
        if url.endswith("/") and not start_key.endswith("/"):
            start_key += "/"

        end_key = inc_last_char(start_key)

    elif match_type == "host":
        if surt_ordered:
            host = start_key.split(")/")[0]

            start_key = host + ")/"
            end_key = host + "*"
        else:
            host = urlparse.urlsplit(url).netloc

            start_key = host + "/"
            end_key = host + "0"

    elif match_type == "domain":
        if not surt_ordered:
            msg = "matchType=domain unsupported for non-surt"
            raise UrlCanonicalizeException(msg)

        host = start_key.split(")/")[0]

        # if tld, use com, as start_key
        # otherwise, stick with com,example)/
        if "," not in host:
            start_key = host + ","
        else:
            start_key = host + ")/"

        # begin brozzler changes
        end_key = host + "~"
        # end brozzler changes
    else:
        raise UrlCanonicalizeException("Invalid match_type: " + match_type)

    return (start_key, end_key)


def monkey_patch_calc_search_range():
    pywb.utils.canonicalize.calc_search_range = _calc_search_range
    pywb.cdx.query.calc_search_range = _calc_search_range


def main(argv=sys.argv):
    brozzler.pywb.TheGoodUrlCanonicalizer.replace_default_canonicalizer()
    brozzler.pywb.TheGoodUrlCanonicalizer.monkey_patch_dsrules_init()
    brozzler.pywb.support_in_progress_warcs()
    brozzler.pywb.monkey_patch_wburl()
    brozzler.pywb.monkey_patch_fuzzy_query()
    brozzler.pywb.monkey_patch_calc_search_range()
    wayback_cli = BrozzlerWaybackCli(
        args=argv[1:],
        default_port=8880,
        desc=("brozzler-wayback - pywb wayback (monkey-patched for use with brozzler)"),
    )
    wayback_cli.run()



"""
brozzler/browser.py - manages the browsers for brozzler

Copyright (C) 2014-2024 Internet Archive

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import base64
import datetime
import json
import logging
import socket
import threading
import time
from ipaddress import AddressValueError

import structlog
import urlcanon
import websocket
from requests.structures import CaseInsensitiveDict

import brozzler
from brozzler.chrome import Chrome


class BrowsingException(Exception):
    pass


class NoBrowsersAvailable(Exception):
    pass


class BrowsingTimeout(BrowsingException):
    pass


class BrowserPool:
    """
    Manages pool of browsers. Automatically chooses available port for the
    debugging protocol.
    """

    logger = structlog.get_logger(logger_name=__module__ + "." + __qualname__)

    def __init__(self, size=3, **kwargs):
        """
        Initializes the pool.

        Args:
            size: size of pool (default 3)
            **kwargs: arguments for Browser(...)
        """
        self.size = size
        self.kwargs = kwargs
        self._in_use = set()
        self._lock = threading.Lock()

    def _fresh_browser(self):
        # choose available port
        sock = socket.socket()
        sock.bind(("0.0.0.0", 0))
        port = sock.getsockname()[1]
        sock.close()

        browser = Browser(port=port, **self.kwargs)
        return browser

    def acquire_multi(self, n=1):
        """
        Returns a list of up to `n` browsers.

        Raises:
            NoBrowsersAvailable if none available
        """
        browsers = []
        with self._lock:
            if len(self._in_use) >= self.size:
                raise NoBrowsersAvailable
            while len(self._in_use) < self.size and len(browsers) < n:
                browser = self._fresh_browser()
                browsers.append(browser)
                self._in_use.add(browser)
        return browsers

    def acquire(self):
        """
        Returns an available instance.

        Returns:
            browser from pool, if available

        Raises:
            NoBrowsersAvailable if none available
        """
        with self._lock:
            if len(self._in_use) >= self.size:
                raise NoBrowsersAvailable
            browser = self._fresh_browser()
            self._in_use.add(browser)
            return browser

    def release(self, browser):
        browser.stop()  # make sure
        with self._lock:
            self._in_use.remove(browser)

    def release_all(self, browsers):
        for browser in browsers:
            browser.stop()  # make sure
        with self._lock:
            for browser in browsers:
                self._in_use.remove(browser)

    def shutdown_now(self):
        self.logger.info(
            "shutting down browser pool (%s browsers in use)", len(self._in_use)
        )
        with self._lock:
            for browser in self._in_use:
                browser.stop()

    def num_available(self):
        return self.size - len(self._in_use)

    def num_in_use(self):
        return len(self._in_use)


# uncomment the next line for LOTS of debugging logging
# websocket.enableTrace(True)


class WebsockReceiverThread(threading.Thread):
    logger = structlog.get_logger(logger_name=__module__ + "." + __qualname__)

    def __init__(self, websock, name=None, daemon=True):
        super().__init__(name=name, daemon=daemon)

        self.websock = websock

        self.calling_thread = threading.current_thread()

        self.websock.on_open = self._on_open
        self.websock.on_message = self._on_message
        self.websock.on_error = self._on_error
        self.websock.on_close = self._on_close

        self.is_open = False
        self.got_page_load_event = None
        self.page_status = None  # Loaded page HTTP status code
        self.reached_limit = None

        self.on_request = None
        self.on_response = None
        self.on_service_worker_version_updated = None

        self.activity_lock = threading.Lock()
        self.last_network_activity = time.time()  # Latest time a request finished
        self.active_connections = set()

        self.initial_document = None

        self._result_messages = {}

    def expect_result(self, msg_id):
        self._result_messages[msg_id] = None

    def received_result(self, msg_id):
        return bool(self._result_messages.get(msg_id))

    def pop_result(self, msg_id):
        return self._result_messages.pop(msg_id)

    def _on_close(self, websock, close_status_code, close_msg):
        pass
        # self.logger.info('GOODBYE GOODBYE WEBSOCKET')

    def _on_open(self, websock):
        self.is_open = True

    def _on_error(self, websock, e):
        """
        Raises BrowsingException in the thread that created this instance.
        """
        if isinstance(
            e, (websocket.WebSocketConnectionClosedException, ConnectionResetError)
        ):
            self.logger.error("websocket closed, did chrome die?")
        else:
            self.logger.exception("exception from websocket receiver thread")
        brozzler.thread_raise(self.calling_thread, BrowsingException)

    def run(self):
        # ping_timeout is used as the timeout for the call to select.select()
        # in addition to its documented purpose, and must have a value to avoid
        # hangs in certain situations
        #
        # skip_ut8_validation is a recommended performance improvement:
        # https://websocket-client.readthedocs.io/en/latest/faq.html#why-is-this-library-slow
        self.websock.run_forever(
            sockopt=((socket.IPPROTO_TCP, socket.TCP_NODELAY, 1),),
            ping_timeout=0.5,
            skip_utf8_validation=True,
        )

    def _on_message(self, websock, message):
        try:
            self._handle_message(websock, message)
        except:  # noqa: E722
            self.logger.exception(
                "uncaught exception in _handle_message",
                message=message,
            )

    def _network_response_received(self, message):
        status = message["params"]["response"].get("status")
        if status == 420 and "Warcprox-Meta" in CaseInsensitiveDict(
            message["params"]["response"]["headers"]
        ):
            if not self.reached_limit:
                warcprox_meta = json.loads(
                    CaseInsensitiveDict(message["params"]["response"]["headers"])[
                        "Warcprox-Meta"
                    ]
                )
                self.reached_limit = brozzler.ReachedLimit(warcprox_meta=warcprox_meta)
                self.logger.info("reached limit", limit=self.reached_limit)
                brozzler.thread_raise(self.calling_thread, brozzler.ReachedLimit)
            else:
                self.logger.info(
                    "reached limit but self.reached_limit is already set, "
                    "assuming the calling thread is already handling this"
                )
        if self.on_response:
            self.on_response(message)

        if status and self.page_status is None:
            self.page_status = status

    def _javascript_dialog_opening(self, message):
        self.logger.info("javascript dialog opened", message=message)
        if message["params"]["type"] == "alert":
            accept = True
        else:
            accept = False
        self.websock.send(
            json.dumps(
                dict(
                    id=0,
                    method="Page.handleJavaScriptDialog",
                    params={"accept": accept},
                ),
                separators=(",", ":"),
            )
        )

    def _should_track_request(self, message) -> bool:
        """
        Decides whether or not to include a request in the idle check.
        """
        # Workaround for https://github.com/GoogleChrome/lighthouse/issues/11850
        # Don't add frame URLs to the set of active requests because they will
        # never receive a loadingFinished event.
        if (
            "params" in message
            and "type" in message["params"]
            and "frameId" in message["params"]
            and message["params"]["type"] == "Document"
        ):
            if self.initial_document is None:
                self.initial_document = message["params"]["frameId"]
            return self.initial_document == message["params"]["frameId"]
        return True

    def _handle_message(self, websock, json_message):
        message = json.loads(json_message)
        if "method" in message:
            if message["method"] == "Page.loadEventFired":
                self.got_page_load_event = datetime.datetime.utcnow()
            elif message["method"] == "Network.responseReceived":
                self._network_response_received(message)
                with self.activity_lock:
                    self.last_network_activity = time.time()
            elif message["method"] == "Network.requestWillBeSent":
                if self.on_request:
                    self.on_request(message)

                if "params" in message and "requestId" in message["params"]:
                    with self.activity_lock:
                        if self._should_track_request(message):
                            self.active_connections.add(message["params"]["requestId"])
                        self.last_network_activity = time.time()
            elif (
                message["method"] == "Network.dataReceived"
                and "params" in message
                and "requestId" in message["params"]
            ):
                with self.activity_lock:
                    self.last_network_activity = time.time()
            elif (
                message["method"] == "Network.loadingFinished"
                and "params" in message
                and "requestId" in message["params"]
            ):
                with self.activity_lock:
                    self.active_connections.discard(message["params"]["requestId"])
                    self.last_network_activity = time.time()
            elif message["method"] == "Network.loadingFailed" and "params" in message:
                if "requestId" in message["params"]:
                    with self.activity_lock:
                        self.active_connections.discard(message["params"]["requestId"])
                        self.last_network_activity = time.time()
                if (
                    "errorText" in message["params"]
                    and message["params"]["errorText"]
                    == "net::ERR_PROXY_CONNECTION_FAILED"
                ):
                    brozzler.thread_raise(self.calling_thread, brozzler.ProxyError)
            elif message["method"] == "Page.interstitialShown":
                # AITFIVE-1529: handle http auth
                # we should kill the browser when we receive Page.interstitialShown and
                # consider the page finished, until this is fixed:
                # https://bugs.chromium.org/p/chromium/issues/detail?id=764505
                self.logger.info(
                    "Page.interstialShown (likely unsupported http auth request)"
                )
                brozzler.thread_raise(
                    self.calling_thread, brozzler.PageInterstitialShown
                )
            elif message["method"] == "Inspector.targetCrashed":
                self.logger.error("""chrome tab went "aw snap" or "he's dead jim"!""")
                brozzler.thread_raise(self.calling_thread, BrowsingException)
            elif message["method"] == "Console.messageAdded":
                self.logger.debug(
                    "console.%s %s",
                    message["params"]["message"]["level"],
                    message["params"]["message"]["text"],
                )
            elif message["method"] == "Runtime.exceptionThrown":
                self.logger.debug("uncaught exception", message=message)
            elif message["method"] == "Page.javascriptDialogOpening":
                self._javascript_dialog_opening(message)
            elif message["method"] == "ServiceWorker.workerVersionUpdated":
                if self.on_service_worker_version_updated:
                    self.on_service_worker_version_updated(message)
            # else:
            #     self.logger.debug("%s %s", message["method"], json_message)
        elif "result" in message:
            if message["id"] in self._result_messages:
                self._result_messages[message["id"]] = message

    #      else:
    #          self.logger.debug("%s", json_message)
    #  else:
    #      self.logger.debug("%s", json_message)


class Browser:
    """
    Manages an instance of Chrome for browsing pages.
    """

    logger = structlog.get_logger(logger_name=__module__ + "." + __qualname__)

    def __init__(self, **kwargs):
        """
        Initializes the Browser.

        Args:
            **kwargs: arguments for Chrome(...)
        """
        self.chrome = Chrome(**kwargs)
        self.websock_url = None
        self.websock = None
        self.websock_thread = None
        self.is_browsing = False
        self._command_id = Counter()
        self._wait_interval = 0.5
        self._max_screenshot_width = kwargs.get("max_screenshot_width", 2000)
        self._max_screenshot_height = kwargs.get("max_screenshot_height", 20000)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *args):
        self.stop()

    def _wait_for(self, callback, timeout=None):
        """
        Spins until callback() returns truthy.
        """
        start = time.time()
        while True:
            if callback():
                return
            elapsed = time.time() - start
            if timeout and elapsed > timeout:
                raise BrowsingTimeout(
                    "timed out after %.1fs waiting for: %s" % (elapsed, callback)
                )
            brozzler.sleep(self._wait_interval)

    def send_to_chrome(self, suppress_logging=False, **kwargs):
        msg_id = next(self._command_id)
        kwargs["id"] = msg_id
        msg = json.dumps(kwargs, separators=(",", ":"))
        self.logger.debug(
            "sending message",
            websock=self.websock,
            message=msg,
        )
        self.websock.send(msg)
        return msg_id

    def start(self, **kwargs):
        """
        Starts chrome if it's not running.

        Args:
            **kwargs: arguments for self.chrome.start(...)
        """
        if not self.is_running():
            self.websock_url = self.chrome.start(**kwargs)
            self.websock = websocket.WebSocketApp(self.websock_url)
            self.websock_thread = WebsockReceiverThread(
                self.websock, name="WebsockThread:%s" % self.chrome.port
            )
            self.websock_thread.start()

            self._wait_for(lambda: self.websock_thread.is_open, timeout=30)

            # tell browser to send us messages we're interested in
            self.send_to_chrome(method="Network.enable")
            self.send_to_chrome(method="Page.enable")
            # Enable Console & Runtime output only when debugging.
            # After all, we just print these events with debug(), we don't use
            # them in Brozzler logic.
            if self.logger.is_enabled_for(logging.DEBUG):
                self.send_to_chrome(method="Console.enable")
                self.send_to_chrome(method="Runtime.enable")
            self.send_to_chrome(method="ServiceWorker.enable")
            self.send_to_chrome(method="ServiceWorker.setForceUpdateOnPageLoad")

            # disable google analytics and amp analytics
            self.send_to_chrome(
                method="Network.setBlockedURLs",
                params={
                    "urls": [
                        "*google-analytics.com/analytics.js*",
                        "*google-analytics.com/ga.js*",
                        "*google-analytics.com/ga_exp.js*",
                        "*google-analytics.com/urchin.js*",
                        "*google-analytics.com/collect*",
                        "*google-analytics.com/r/collect*",
                        "*google-analytics.com/__utm.gif*",
                        "*google-analytics.com/gtm/js?*",
                        "*google-analytics.com/cx/api.js*",
                        "*cdn.ampproject.org/*/amp-analytics*.js",
                    ]
                },
            )

    def stop(self):
        """
        Stops chrome if it's running.
        """
        try:
            if self.websock and self.websock.sock and self.websock.sock.connected:
                self.logger.info("shutting down websocket connection")
                try:
                    self.websock.close()
                except BaseException:
                    self.logger.exception(
                        "exception closing websocket", websocket=self.websock
                    )

            self.chrome.stop()

            if self.websock_thread and (
                self.websock_thread != threading.current_thread()
            ):
                self.websock_thread.join(timeout=30)
                if self.websock_thread.is_alive():
                    self.logger.error(
                        "%s still alive 30 seconds after closing %s, will "
                        "forcefully nudge it again",
                        self.websock_thread,
                        self.websock,
                    )
                    self.websock.keep_running = False
                    self.websock_thread.join(timeout=30)
                    if self.websock_thread.is_alive():
                        self.logger.critical(
                            "%s still alive 60 seconds after closing %s",
                            self.websock_thread,
                            self.websock,
                        )

            self.websock_url = None
        except:  # noqa: E722
            self.logger.exception("problem stopping")

    def is_running(self):
        return self.websock_url is not None

    def browse_page(
        self,
        page_url,
        extra_headers=None,
        user_agent=None,
        behavior_parameters=None,
        behaviors_dir=None,
        on_request=None,
        on_response=None,
        on_service_worker_version_updated=None,
        on_screenshot=None,
        username=None,
        password=None,
        hashtags=None,
        screenshot_full_page=False,
        skip_extract_outlinks=False,
        skip_visit_hashtags=False,
        skip_youtube_dl=False,
        ytdlp_tmpdir="/tmp",
        simpler404=False,
        page_timeout=300,
        behavior_timeout=300,
        extract_outlinks_timeout=60,
        download_throughput=-1,
        stealth=False,
    ):
        """
        Browses page in browser.

        Browser should already be running, i.e. start() should have been
        called. Opens the page_url in the browser, runs behaviors, takes a
        screenshot, extracts outlinks.

        Args:
            page_url: url of the page to browse
            extra_headers: dict of extra http headers to configure the browser
                to send with every request (default None)
            user_agent: user agent string, replaces browser default if
                supplied (default None)
            behavior_parameters: dict of parameters for populating the
                javascript behavior template (default None)
            behaviors_dir: Directory containing behaviors.yaml and JS templates
                (default None loads Brozzler default JS behaviors)
            on_request: callback to invoke on every Network.requestWillBeSent
                event, takes one argument, the json-decoded message (default
                None)
            on_response: callback to invoke on every Network.responseReceived
                event, takes one argument, the json-decoded message (default
                None)
            on_service_worker_version_updated: callback to invoke on every
                ServiceWorker.workerVersionUpdated event, takes one argument,
                the json-decoded message (default None)
            on_screenshot: callback to invoke when screenshot is obtained,
                takes one argument, the the raw jpeg bytes (default None)
                # XXX takes two arguments, the url of the page at the time the
                # screenshot was taken, and the raw jpeg bytes (default None)
            username: username string to use to try logging in if a login form
                is found in the page (default None)
            password: password string to use to try logging in if a login form
                is found in the page (default None)
            ... (there are more)

        Returns:
            A tuple (final_page_url, outlinks).
            final_page_url: the url in the location bar at the end of the
                browse_page cycle, which could be different from the original
                page url if the page redirects, javascript has changed the url
                in the location bar, etc
            outlinks: a list of navigational links extracted from the page

        Raises:
            brozzler.ProxyError: in case of proxy connection error
            BrowsingException: if browsing the page fails in some other way
        """
        if not self.is_running():
            raise BrowsingException("browser has not been started")
        if self.is_browsing:
            raise BrowsingException("browser is already busy browsing a page")
        self.is_browsing = True
        if on_request:
            self.websock_thread.on_request = on_request
        if on_response:
            self.websock_thread.on_response = on_response
        if on_service_worker_version_updated:
            self.websock_thread.on_service_worker_version_updated = (
                on_service_worker_version_updated
            )
        try:
            with brozzler.thread_accept_exceptions():
                self.configure_browser(
                    extra_headers=extra_headers,
                    user_agent=user_agent,
                    download_throughput=download_throughput,
                    stealth=stealth,
                )
                self.navigate_to_page(page_url, timeout=page_timeout)
                if password:
                    self.try_login(username, password, timeout=page_timeout)
                    # if login redirected us, return to page_url
                    if page_url != self.url().split("#")[0]:
                        self.logger.debug(
                            "login navigated away; returning!", page_url=page_url
                        )
                        self.navigate_to_page(page_url, timeout=page_timeout)
                    # allow time for any additional redirects or scripts to run after login
                    time.sleep(10)
                    self._wait_for_idle(idle_time=5, timeout=4)
                # If the target page HTTP status is 4xx/5xx, there is no point
                # in running behaviors, screenshot, outlink and hashtag
                # extraction as we didn't get a valid page.
                # This is only enabled with option `simpler404`.
                run_behaviors = True
                if simpler404 and (
                    self.websock_thread.page_status is None
                    or self.websock_thread.page_status >= 400
                ):
                    run_behaviors = False

                behavior_outlinks: frozenset[str] = frozenset()
                if run_behaviors and behavior_timeout > 0:
                    behavior_script = brozzler.behavior_script(
                        page_url, behavior_parameters, behaviors_dir=behaviors_dir
                    )
                    behavior_outlinks = self.run_behavior(
                        behavior_script, timeout=behavior_timeout
                    )
                final_page_url = self.url()
                if on_screenshot:
                    if simpler404:
                        if (
                            self.websock_thread.page_status
                            and self.websock_thread.page_status < 400
                        ):
                            self._try_screenshot(on_screenshot, screenshot_full_page)
                    else:
                        self._try_screenshot(on_screenshot, screenshot_full_page)

                if not run_behaviors or skip_extract_outlinks:
                    outlinks: frozenset[str] = frozenset()
                else:
                    outlinks = self.extract_outlinks(timeout=extract_outlinks_timeout)
                if run_behaviors and not skip_visit_hashtags:
                    self.visit_hashtags(final_page_url, hashtags, outlinks)
                return final_page_url, outlinks.union(behavior_outlinks)
        except brozzler.ReachedLimit:
            # websock_thread has stashed the ReachedLimit exception with
            # more information, raise that one
            raise self.websock_thread.reached_limit
        except websocket.WebSocketConnectionClosedException as e:
            self.logger.exception("websocket closed, did chrome die?")
            raise BrowsingException(e)
        finally:
            self.is_browsing = False
            self.websock_thread.on_request = None
            self.websock_thread.on_response = None

    def _try_screenshot(self, on_screenshot, full_page=False):
        """The browser instance must be scrolled to the top of the page before
        trying to get a screenshot.
        """
        self.send_to_chrome(
            method="Runtime.evaluate",
            suppress_logging=True,
            params={"expression": "window.scroll(0,0)"},
        )
        for i in range(3):
            try:
                jpeg_bytes = self.screenshot(full_page)
                on_screenshot(jpeg_bytes)
                return
            except BrowsingTimeout:
                self.logger.exception("attempt %s/3", i + 1)

    def visit_hashtags(self, page_url, hashtags, outlinks):
        _hashtags = set(hashtags or [])
        for outlink in outlinks:
            url = urlcanon.whatwg(outlink)
            hashtag = (url.hash_sign + url.fragment).decode("utf-8")
            urlcanon.canon.remove_fragment(url)
            if hashtag and str(url) == page_url:
                _hashtags.add(hashtag)
        # could inject a script that listens for HashChangeEvent to figure
        # out which hashtags were visited already and skip those
        for hashtag in _hashtags:
            # navigate_to_hashtag (nothing to wait for so no timeout?)
            self.logger.debug("navigating to hashtag", hashtag=hashtag)
            url = urlcanon.whatwg(page_url)
            url.hash_sign = b"#"
            url.fragment = hashtag[1:].encode("utf-8")
            self.send_to_chrome(method="Page.navigate", params={"url": str(url)})
            time.sleep(1)  # allow time for any scripts to run
            self._wait_for_idle(idle_time=2, timeout=4)
            # take another screenshot?
            # run behavior again with short timeout?
            # retrieve outlinks again and append to list?

    def _wait_for_idle(self, idle_time: float, timeout: float):
        """
        Waits up to timeout seconds for the network to be idle.
        "Idle" is defined as having no active requests for at least idle_time seconds.

        Args:
            idle_time: The minimum required idle time. (This function may not wait at all
                if the network has already been idle for this time.)
            timeout: The maximum number of seconds to wait for.
        """

        def is_idle():
            with self.websock_thread.activity_lock:
                elapsed = time.time() - self.websock_thread.last_network_activity
                return (
                    len(self.websock_thread.active_connections) == 0
                    and elapsed > idle_time
                )

        try:
            self._wait_for(is_idle, timeout)
        except BrowsingTimeout:
            self.logger.debug("idle timed out")

    def configure_browser(
        self, extra_headers=None, user_agent=None, download_throughput=-1, stealth=False
    ):
        headers = extra_headers or {}
        headers["Accept-Encoding"] = "gzip"  # avoid encodings br, sdch
        msg_id = self.send_to_chrome(
            method="Network.setExtraHTTPHeaders", params={"headers": headers}
        )
        if user_agent:
            msg_id = self.send_to_chrome(
                method="Network.setUserAgentOverride", params={"userAgent": user_agent}
            )
        if download_throughput > -1:
            # traffic shaping already used by SPN2 to aid warcprox resilience
            # parameter value as bytes/second, or -1 to disable (default)
            msg_id = self.send_to_chrome(
                method="Network.emulateNetworkConditions",
                params={"downloadThroughput": download_throughput},
            )
        if stealth:
            self.websock_thread.expect_result(self._command_id.peek())
            js = brozzler.jinja2_environment().get_template("stealth.js").render()
            msg_id = self.send_to_chrome(
                method="Page.addScriptToEvaluateOnNewDocument", params={"source": js}
            )
            self._wait_for(
                lambda: self.websock_thread.received_result(msg_id), timeout=10
            )

    def navigate_to_page(self, page_url, timeout=300):
        self.logger.info("navigating to page", page_url=page_url)
        self.websock_thread.got_page_load_event = None
        self.websock_thread.page_status = None
        self.send_to_chrome(method="Page.navigate", params={"url": page_url})
        self._wait_for(lambda: self.websock_thread.got_page_load_event, timeout=timeout)

    def inject_outlink_extractor(self, timeout=60):
        self.websock_thread.expect_result(self._command_id.peek())
        js = brozzler.jinja2_environment().get_template("extract-outlinks.js").render()
        # This defines the method but doesn't extract outlinks yet
        msg_id = self.send_to_chrome(
            method="Runtime.evaluate",
            suppress_logging=True,
            params={"expression": js},
        )
        self._wait_for(
            lambda: self.websock_thread.received_result(msg_id), timeout=timeout
        )

    def extract_outlinks(self, timeout=60) -> frozenset[str]:
        self.logger.info("extracting outlinks")

        # Define the outlink extractor first before extracting
        self.inject_outlink_extractor(timeout=timeout)

        self.websock_thread.expect_result(self._command_id.peek())
        # Now we actually do outlink extraction
        msg_id = self.send_to_chrome(
            method="Runtime.evaluate",
            params={
                "expression": "__brzl_extractOutlinks()",
                # returnByValue ensures we can receive an array response
                "returnByValue": True,
            },
        )
        self._wait_for(
            lambda: self.websock_thread.received_result(msg_id), timeout=timeout
        )

        message = self.websock_thread.pop_result(msg_id)
        if (
            "result" in message
            and "result" in message["result"]
            and "value" in message["result"]["result"]
        ):
            if message["result"]["result"]["value"]:
                out = []
                for link in message["result"]["result"]["value"]:
                    try:
                        out.append(str(urlcanon.whatwg(link)))
                    except AddressValueError:
                        self.logger.warning("skip invalid outlink", outlink=link)
                return frozenset(out)
            else:
                # no links found
                return frozenset()
        else:
            self.logger.error("problem extracting outlinks", message=message)
            return frozenset()

    def screenshot(self, full_page=False, timeout=45):
        """Optionally capture full page screenshot using puppeteer as an
        inspiration:
        https://github.com/GoogleChrome/puppeteer/blob/master/lib/Page.js#L898
        """
        self.logger.info("taking screenshot")
        if full_page:
            self.websock_thread.expect_result(self._command_id.peek())
            msg_id = self.send_to_chrome(method="Page.getLayoutMetrics")
            self._wait_for(
                lambda: self.websock_thread.received_result(msg_id), timeout=timeout
            )
            message = self.websock_thread.pop_result(msg_id)
            width = min(
                message["result"]["contentSize"]["width"], self._max_screenshot_width
            )
            height = min(
                message["result"]["contentSize"]["height"], self._max_screenshot_height
            )
            clip = dict(x=0, y=0, width=width, height=height, scale=1)
            deviceScaleFactor = 1
            screenOrientation = {"angle": 0, "type": "portraitPrimary"}
            self.send_to_chrome(
                method="Emulation.setDeviceMetricsOverride",
                params=dict(
                    mobile=False,
                    width=width,
                    height=height,
                    deviceScaleFactor=deviceScaleFactor,
                    screenOrientation=screenOrientation,
                ),
            )
            capture_params = {"format": "jpeg", "quality": 95, "clip": clip}
        else:
            capture_params = {"format": "jpeg", "quality": 95}
        self.websock_thread.expect_result(self._command_id.peek())
        msg_id = self.send_to_chrome(
            method="Page.captureScreenshot", params=capture_params
        )
        self._wait_for(
            lambda: self.websock_thread.received_result(msg_id), timeout=timeout
        )
        message = self.websock_thread.pop_result(msg_id)
        jpeg_bytes = base64.b64decode(message["result"]["data"])
        return jpeg_bytes

    def url(self, timeout=30):
        """
        Returns value of document.URL from the browser.
        """
        self.websock_thread.expect_result(self._command_id.peek())
        msg_id = self.send_to_chrome(
            method="Runtime.evaluate", params={"expression": "document.URL"}
        )
        self._wait_for(
            lambda: self.websock_thread.received_result(msg_id), timeout=timeout
        )
        message = self.websock_thread.pop_result(msg_id)
        return message["result"]["result"]["value"]

    def run_behavior(self, behavior_script, timeout=900) -> frozenset[str]:
        # Inject the outlink extractor so it's available to behaviors
        self.inject_outlink_extractor(timeout=timeout)

        self.send_to_chrome(
            method="Runtime.evaluate",
            suppress_logging=True,
            params={"expression": behavior_script},
        )

        check_interval = min(timeout, 7)
        start = time.time()
        while True:
            elapsed = time.time() - start
            if elapsed > timeout:
                self.logger.info("behavior reached hard timeout", elapsed=elapsed)
                return frozenset()

            brozzler.sleep(check_interval)

            self.websock_thread.expect_result(self._command_id.peek())
            msg_id = self.send_to_chrome(
                method="Runtime.evaluate",
                suppress_logging=True,
                params={
                    "expression": "umbraBehaviorFinished()",
                    # returnByValue ensures we can return more complicated types like dicts
                    "returnByValue": True,
                },
            )
            try:
                self._wait_for(
                    lambda: self.websock_thread.received_result(msg_id), timeout=5
                )
                msg = self.websock_thread.pop_result(msg_id)
                if (
                    msg
                    and "result" in msg
                    and "exceptionDetails" not in msg["result"]
                    and not (
                        "wasThrown" in msg["result"] and msg["result"]["wasThrown"]
                    )
                    and "result" in msg["result"]
                ):
                    if isinstance(msg["result"]["result"]["value"], bool):
                        if msg["result"]["result"]["value"]:
                            self.logger.info("behavior decided it has finished")
                            return frozenset()
                    # new-style response dict that has more than just a finished bool
                    elif isinstance(msg["result"]["result"]["value"], dict):
                        response = msg["result"]["result"]["value"]
                        if response["finished"]:
                            self.logger.info("behavior decided it has finished")
                            outlinks = frozenset(response.get("outlinks", []))
                            return outlinks
            except BrowsingTimeout:
                pass

    def try_login(self, username, password, timeout=300):
        try_login_js = (
            brozzler.jinja2_environment()
            .get_template("try-login.js.j2")
            .render(username=username, password=password)
        )

        self.logger.info("trying to login")
        time.sleep(1)  # allow time for any scripts to run
        self._wait_for_idle(idle_time=2, timeout=4)

        self.websock_thread.got_page_load_event = None
        self.send_to_chrome(
            method="Runtime.evaluate",
            suppress_logging=True,
            params={"expression": try_login_js},
        )

        # wait for tryLogin to finish trying (should be very very quick)
        start = time.time()
        while True:
            self.websock_thread.expect_result(self._command_id.peek())
            msg_id = self.send_to_chrome(
                method="Runtime.evaluate",
                params={
                    "expression": 'try { __brzl_tryLoginState } catch (e) { "maybe-submitted-form" }'
                },
            )
            try:
                self._wait_for(
                    lambda: self.websock_thread.received_result(msg_id), timeout=5
                )
                msg = self.websock_thread.pop_result(msg_id)
                if msg and "result" in msg and "result" in msg["result"]:
                    result = msg["result"]["result"]["value"]
                    if result == "login-form-not-found":
                        # we're done
                        return
                    elif result in ("submitted-form", "maybe-submitted-form"):
                        # wait for page load event below
                        self.logger.info(
                            "submitted a login form, waiting for another "
                            "page load event"
                        )
                        break
                    # else try again to get __brzl_tryLoginState

            except BrowsingTimeout:
                pass

            if time.time() - start > 30:
                raise BrowsingException(
                    "timed out trying to check if tryLogin finished"
                )

        # if we get here, we submitted a form, now we wait for another page
        # load event
        self._wait_for(lambda: self.websock_thread.got_page_load_event, timeout=timeout)


class Counter:
    def __init__(self):
        self.next_value = 0

    def __next__(self):
        try:
            return self.next_value
        finally:
            self.next_value += 1

    def peek(self):
        return self.next_value



"""
brozzler/chrome.py - manages the chrome/chromium browser for brozzler

Copyright (C) 2014-2023 Internet Archive

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import os
import re
import select
import signal
import sqlite3
import subprocess
import sys
import tempfile
import threading
import time
import urllib.request
from functools import lru_cache

import structlog

import brozzler


@lru_cache
def check_version(chrome_exe):
    """
    Raises SystemExit if `chrome_exe` is not a supported browser version.

    Must run in the main thread to have the desired effect.
    """
    # mac$ /Applications/Google\ Chrome.app/Contents/MacOS/Google\ Chrome --version
    # Google Chrome 64.0.3282.140
    # mac$ /Applications/Google\ Chrome\ Canary.app/Contents/MacOS/Google\ Chrome\ Canary --version
    # Google Chrome 66.0.3341.0 canary
    # linux$ chromium-browser --version
    # Using PPAPI flash.
    #  --ppapi-flash-path=/usr/lib/adobe-flashplugin/libpepflashplayer.so --ppapi-flash-version=
    # Chromium 61.0.3163.100 Built on Ubuntu , running on Ubuntu 16.04
    cmd = [chrome_exe, "--version"]
    out = subprocess.check_output(cmd, timeout=60)
    m = re.search(rb"(Chromium|Google Chrome) ([\d.]+)", out)
    if not m:
        sys.exit(
            "unable to parse browser version from output of "
            "%r: %r" % (subprocess.list2cmdline(cmd), out)
        )
    version_str = m.group(2).decode()
    major_version = int(version_str.split(".")[0])
    if major_version < 64:
        sys.exit(
            "brozzler requires chrome/chromium version 64 or "
            "later but %s reports version %s" % (chrome_exe, version_str)
        )
    return major_version


class Chrome:
    logger = structlog.get_logger(logger_name=__module__ + "." + __qualname__)

    def __init__(self, chrome_exe, port=9222, ignore_cert_errors=False):
        """
        Initializes instance of this class.

        Doesn't start the browser, start() does that.

        Args:
            chrome_exe: filesystem path to chrome/chromium executable
            port: chrome debugging protocol port (default 9222)
            ignore_cert_errors: configure chrome to accept all certs (default
                False)
        """
        self.port = port
        self.chrome_exe = chrome_exe
        self.ignore_cert_errors = ignore_cert_errors
        self._shutdown = threading.Event()
        self.chrome_process = None

    def __enter__(self):
        """
        Returns websocket url to chrome window with about:blank loaded.
        """
        return self.start()

    def __exit__(self, *args):
        self.stop()

    def _init_cookie_db(self, cookie_db):
        cookie_dir = os.path.join(self._chrome_user_data_dir, "Default")
        cookie_location = os.path.join(cookie_dir, "Cookies")
        cookie_logger = self.logger.bind(cookie_location=cookie_location)

        cookie_logger.debug("cookie DB provided, writing to")
        os.makedirs(cookie_dir, exist_ok=True)

        try:
            with open(cookie_location, "wb") as cookie_file:
                cookie_file.write(cookie_db)
        except OSError:
            cookie_logger.exception("exception writing cookie file")

    def persist_and_read_cookie_db(self):
        cookie_location = os.path.join(self._chrome_user_data_dir, "Default", "Cookies")
        cookie_logger = self.logger.bind(cookie_location=cookie_location)
        cookie_logger.debug(
            "marking cookies persistent then reading file into memory",
        )
        try:
            with sqlite3.connect(cookie_location) as conn:
                cur = conn.cursor()
                cur.execute("UPDATE cookies SET is_persistent = 1")
        except sqlite3.Error:
            try:
                # db schema changed around version 66, this is the old schema
                with sqlite3.connect(cookie_location) as conn:
                    cur = conn.cursor()
                    cur.execute("UPDATE cookies SET persistent = 1")
            except sqlite3.Error:
                cookie_logger.exception("exception updating cookie DB")

        cookie_db = None
        try:
            with open(cookie_location, "rb") as cookie_file:
                cookie_db = cookie_file.read()
        except OSError:
            cookie_logger.exception("exception reading from cookie DB file")
        return cookie_db

    def start(
        self,
        proxy=None,
        cookie_db=None,
        disk_cache_dir=None,
        disk_cache_size=None,
        websocket_timeout=60,
        window_height=900,
        window_width=1400,
        headless=True,
    ):
        """
        Starts chrome/chromium process.

        Args:
            proxy: http proxy 'host:port' (default None)
            cookie_db: raw bytes of chrome/chromium sqlite3 cookies database,
                which, if supplied, will be written to
                {chrome_user_data_dir}/Default/Cookies before running the
                browser (default None)
            disk_cache_dir: use directory for disk cache. The default location
                is inside `self._home_tmpdir` (default None).
            disk_cache_size: Forces the maximum disk space to be used by the disk
                cache, in bytes. (default None)
            websocket_timeout: websocket timeout, in seconds
            window_height, window_width: window height and width, in pixels
        Returns:
            websocket url to chrome window with about:blank loaded
        """
        # these can raise exceptions
        self._home_tmpdir = tempfile.TemporaryDirectory()
        self._chrome_user_data_dir = os.path.join(
            self._home_tmpdir.name, "chrome-user-data"
        )
        if cookie_db:
            self._init_cookie_db(cookie_db)
        self._shutdown.clear()

        new_env = os.environ.copy()
        new_env["HOME"] = self._home_tmpdir.name
        chrome_args = [
            self.chrome_exe,
            "-v",
            "--remote-debugging-port=%s" % self.port,
            "--remote-allow-origins=http://localhost:%s" % self.port,
            "--use-mock-keychain",  # mac thing
            "--user-data-dir=%s" % self._chrome_user_data_dir,
            "--disable-background-networking",
            "--disable-breakpad",
            "--disable-renderer-backgrounding",
            "--disable-hang-monitor",
            "--disable-background-timer-throttling",
            "--mute-audio",
            "--disable-web-sockets",
            f"--window-size={window_width},{window_height}",
            "--no-default-browser-check",
            "--disable-first-run-ui",
            "--no-first-run",
            "--homepage=about:blank",
            "--disable-features=HttpsUpgrades,HttpsFirstBalancedModeAutoEnable,OptimizationGuideModelDownloading,OptimizationHintsFetching,OptimizationTargetPrediction,OptimizationHints",
            "--disable-direct-npapi-requests",
            "--disable-web-security",
            "--disable-notifications",
            "--disable-extensions",
            "--disable-save-password-bubble",
            "--disable-sync",
        ]
        if headless:
            major_version = check_version(self.chrome_exe)
            if major_version >= 109:
                chrome_args.append("--headless=new")
            elif 96 <= major_version <= 108:
                chrome_args.append("--headless=chrome")
            else:
                chrome_args.append("--headless")

        extra_chrome_args = os.environ.get("BROZZLER_EXTRA_CHROME_ARGS")
        if extra_chrome_args:
            chrome_args.extend(extra_chrome_args.split())
        if disk_cache_dir:
            chrome_args.append("--disk-cache-dir=%s" % disk_cache_dir)
        if disk_cache_size:
            chrome_args.append("--disk-cache-size=%s" % disk_cache_size)
        if self.ignore_cert_errors:
            chrome_args.append("--ignore-certificate-errors")
        if proxy:
            chrome_args.append("--proxy-server=%s" % proxy)
        chrome_args.append("about:blank")
        self.logger.info(
            "running", chrome_args=subprocess.list2cmdline(chrome_args), proxy=proxy
        )
        # start_new_session - new process group so we can kill the whole group
        self.chrome_process = subprocess.Popen(
            chrome_args,
            env=new_env,
            start_new_session=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            bufsize=0,
        )
        self._out_reader_thread = threading.Thread(
            target=self._read_stderr_stdout,
            name="ChromeOutReaderThread:%s" % self.port,
            daemon=True,
        )
        self._out_reader_thread.start()
        self.logger.info("chrome running", pid=self.chrome_process.pid)

        return self._websocket_url(timeout_sec=websocket_timeout)

    def _websocket_url(self, timeout_sec=60):
        json_url = "http://localhost:%s/json" % self.port
        url_logger = self.logger.bind(json_url=json_url)
        # make this a member variable so that kill -QUIT reports it
        self._start = time.time()
        self._last_warning = self._start
        while True:
            try:
                raw_json = urllib.request.urlopen(json_url, timeout=30).read()
                all_debug_info = json.loads(raw_json.decode("utf-8"))
                debug_info = [x for x in all_debug_info if x["url"] == "about:blank"]

                if debug_info and "webSocketDebuggerUrl" in debug_info[0]:
                    url_logger.debug("webSocketDebuggerUrl returned", raw_json=raw_json)
                    url = debug_info[0]["webSocketDebuggerUrl"]
                    url_logger.info(
                        "got chrome window websocket debug url",
                        debug_url=url,
                    )
                    return url
            except brozzler.ShutdownRequested:
                raise
            except Exception:
                if time.time() - self._last_warning > 30:
                    url_logger.warning(
                        "problem accessing url (will keep trying until timeout)",
                        timeout_sec=timeout_sec,
                        exc_info=True,
                    )
                    self._last_warning = time.time()
            finally:
                e = None
                if self.chrome_process:
                    if time.time() - self._start > timeout_sec:
                        e = Exception(
                            "killing chrome, failed to retrieve %s after "
                            "%s seconds" % (json_url, time.time() - self._start)
                        )
                    elif self.chrome_process.poll() is not None:
                        e = Exception(
                            "chrome process died with status %s"
                            % self.chrome_process.poll()
                        )
                    else:
                        time.sleep(0.5)
                else:
                    e = Exception("??? self.chrome_process is not set ???")
                if e:
                    self.stop()
                    raise e

    def _read_stderr_stdout(self):
        # XXX select doesn't work on windows
        def readline_nonblock(f):
            buf = b""
            try:
                while (
                    not self._shutdown.is_set()
                    and (len(buf) == 0 or buf[-1] != 0xA)
                    and select.select([f], [], [], 0.5)[0]
                ):
                    buf += f.read(1)
            except (ValueError, OSError):
                # When the chrome process crashes, stdout & stderr are closed
                # and trying to read from them raises these exceptions. We just
                # stop reading and return current `buf`.
                pass
            return buf

        try:
            while not self._shutdown.is_set():
                buf = readline_nonblock(self.chrome_process.stdout)
                if buf:
                    self.logger.debug(
                        "chrome pid %s STDOUT %s", self.chrome_process.pid, buf
                    )

                buf = readline_nonblock(self.chrome_process.stderr)
                if buf:
                    self.logger.debug(
                        "chrome pid %s STDERR %s", self.chrome_process.pid, buf
                    )
        except:  # noqa: E722
            self.logger.exception("unexpected exception")

    def stop(self):
        if not self.chrome_process or self._shutdown.is_set():
            return
        self._shutdown.set()

        pid_logger = self.logger.bind(pid=self.chrome_process.pid)

        timeout_sec = 300
        if self.chrome_process.poll() is None:
            pid_logger.info("terminating chrome")

            os.killpg(self.chrome_process.pid, signal.SIGTERM)
        t0 = time.time()

        try:
            while time.time() - t0 < timeout_sec:
                status = self.chrome_process.poll()
                if status is not None:
                    if status == 0:
                        pid_logger.info("chrome exited normally")
                    else:
                        pid_logger.warning(
                            "chrome exited with nonzero status",
                            status=status,
                        )

                    # XXX I would like to forcefully kill the process group
                    # here to guarantee no orphaned chromium subprocesses hang
                    # around, but there's a chance I suppose that some other
                    # process could have started with the same pgid
                    return
                time.sleep(0.5)

            pid_logger.warning(
                "chrome still alive %.1f seconds after sending "
                "SIGTERM, sending SIGKILL",
                time.time() - t0,
            )
            os.killpg(self.chrome_process.pid, signal.SIGKILL)
            status = self.chrome_process.wait()
            pid_logger.warning(
                "chrome reaped after killing with SIGKILL",
                status=status,
            )

        finally:
            self.chrome_process.stdout.close()
            self.chrome_process.stderr.close()
            try:
                self._home_tmpdir.cleanup()
            except:  # noqa: E722
                self.logger.exception(
                    "exception deleting self._home_tmpdir", tmpdir=self._home_tmpdir
                )
            self._out_reader_thread.join()
            self.chrome_process = None



from typing import Optional

try:
    from http_sd_registry.client import (
        Client,
        Env,
        Registration,
        Scheme,
        format_self_target,
    )
    from http_sd_registry.config import ClientConfig
except ImportError:
    # for users without access to http_sd_registry
    http_sd_registry = None


from prometheus_client import Counter, Gauge, Histogram, start_http_server

# fmt: off
brozzler_in_progress_pages = Gauge("brozzler_in_progress_pages", "number of pages currently processing with brozzler")
brozzler_page_processing_duration_seconds = Histogram("brozzler_page_processing_duration_seconds", "time spent processing a page in brozzler")
brozzler_in_progress_headers = Gauge("brozzler_in_progress_headers", "number of headers currently processing with brozzler")
brozzler_header_processing_duration_seconds = Histogram("brozzler_header_processing_duration_seconds", "time spent processing one page's headers in brozzler")
brozzler_in_progress_browses = Gauge("brozzler_in_progress_browses", "number of pages currently browsing with brozzler")
brozzler_browsing_duration_seconds = Histogram("brozzler_browsing_duration_seconds", "time spent browsing a page in brozzler")
brozzler_in_progress_ytdlps = Gauge("brozzler_in_progress_ytdlps", "number of ytdlp sessions currently in progress with brozzler")
brozzler_ytdlp_duration_seconds = Histogram("brozzler_ytdlp_duration_seconds", "time spent running ytdlp for a page in brozzler")
brozzler_pages_crawled = Counter("brozzler_pages_crawled", "number of pages visited by brozzler")
brozzler_outlinks_found = Counter("brozzler_outlinks_found", "number of outlinks found by brozzler")
brozzler_last_page_crawled_time = Gauge("brozzler_last_page_crawled_time", "time of last page visit, in seconds since UNIX epoch")
brozzler_ydl_urls_checked = Counter("brozzler_ydl_urls_checked", "count of urls checked by brozzler yt-dlp")
brozzler_ydl_extract_successes = Counter("brozzler_ydl_extract_successes", "count of extracts completed by brozzler yt-dlp", labelnames=["youtube_host"])
brozzler_ydl_download_successes = Counter("brozzler_ydl_download_successes", "count of downloads completed by brozzler yt-dlp", labelnames=["youtube_host"])
# fmt: on


def register_prom_metrics(
    metrics_port: int = 8888,
    registry_url: Optional[str] = None,
    env: Optional[str] = None,
):
    # Start metrics endpoint for scraping
    start_http_server(metrics_port)

    if registry_url is None:
        return

    if env == "qa":
        env_for_prom = Env.qa
    elif env == "prod":
        env_for_prom = Env.prod
    else:
        env_for_prom = Env.qa

    config = ClientConfig(server_url_base=registry_url)
    client = Client(config)
    target = format_self_target(scrape_port=metrics_port)
    registration = Registration(
        target=target,
        env=env_for_prom,
        scheme=Scheme.http,
    )
    client.keep_registered_threaded(registration)



#!/usr/bin/env python
"""
test_units.py - some unit tests for parts of brozzler amenable to that

Copyright (C) 2016-2017 Internet Archive

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import datetime
import http.server
import os
import socket
import tempfile
import threading
import time
import uuid
from unittest import mock

import pytest
import yaml

import brozzler
import brozzler.chrome
import brozzler.ydl


@pytest.fixture(scope="module")
def httpd(request):
    # SimpleHTTPRequestHandler always uses CWD so we have to chdir
    os.chdir(os.path.join(os.path.dirname(__file__), "htdocs"))

    httpd = http.server.HTTPServer(
        ("localhost", 0), http.server.SimpleHTTPRequestHandler
    )
    httpd_thread = threading.Thread(name="httpd", target=httpd.serve_forever)
    httpd_thread.start()

    def fin():
        httpd.shutdown()
        httpd.server_close()
        httpd_thread.join()

    request.addfinalizer(fin)

    return httpd


def test_robots(httpd):
    """
    Basic test of robots.txt user-agent substring matching.
    """
    url = "http://localhost:%s/" % httpd.server_port
    site = brozzler.Site(None, {"seed": url, "user_agent": "im/a/GoOdbot/yep"})
    assert brozzler.is_permitted_by_robots(site, url)

    site = brozzler.Site(None, {"seed": url, "user_agent": "im/a bAdBOt/uh huh"})
    assert not brozzler.is_permitted_by_robots(site, url)


def test_robots_http_statuses():
    for status in (
        200,
        204,
        400,
        401,
        402,
        403,
        404,
        405,
        500,
        501,
        502,
        503,
        504,
        505,
    ):

        class Handler(http.server.BaseHTTPRequestHandler):
            def do_GET(self):
                response = (
                    (
                        "HTTP/1.1 %s Meaningless message\r\n"
                        + "Content-length: 0\r\n"
                        + "\r\n"
                    )
                    % status
                ).encode("utf-8")
                self.connection.sendall(response)
                # self.send_response(status)
                # self.end_headers()

        httpd = http.server.HTTPServer(("localhost", 0), Handler)
        httpd_thread = threading.Thread(name="httpd", target=httpd.serve_forever)
        httpd_thread.start()

        try:
            url = "http://localhost:%s/" % httpd.server_port
            site = brozzler.Site(None, {"seed": url})
            assert brozzler.is_permitted_by_robots(site, url)
        finally:
            httpd.shutdown()
            httpd.server_close()
            httpd_thread.join()


def test_robots_empty_response():
    class Handler(http.server.BaseHTTPRequestHandler):
        def do_GET(self):
            self.connection.shutdown(socket.SHUT_RDWR)
            self.connection.close()

    httpd = http.server.HTTPServer(("localhost", 0), Handler)
    httpd_thread = threading.Thread(name="httpd", target=httpd.serve_forever)
    httpd_thread.start()

    try:
        url = "http://localhost:%s/" % httpd.server_port
        site = brozzler.Site(None, {"seed": url})
        assert brozzler.is_permitted_by_robots(site, url)
    finally:
        httpd.shutdown()
        httpd.server_close()
        httpd_thread.join()


def test_robots_socket_timeout():
    stop_hanging = threading.Event()

    class Handler(http.server.BaseHTTPRequestHandler):
        def do_GET(self):
            stop_hanging.wait(60)
            self.connection.sendall(b"HTTP/1.1 200 OK\r\nContent-length: 0\r\n\r\n")

    orig_timeout = brozzler.robots._SessionRaiseOn420.timeout

    httpd = http.server.HTTPServer(("localhost", 0), Handler)
    httpd_thread = threading.Thread(name="httpd", target=httpd.serve_forever)
    httpd_thread.start()

    try:
        url = "http://localhost:%s/" % httpd.server_port
        site = brozzler.Site(None, {"seed": url})
        brozzler.robots._SessionRaiseOn420.timeout = 2
        assert brozzler.is_permitted_by_robots(site, url)
    finally:
        brozzler.robots._SessionRaiseOn420.timeout = orig_timeout
        stop_hanging.set()
        httpd.shutdown()
        httpd.server_close()
        httpd_thread.join()


def test_robots_dns_failure():
    # .invalid. is guaranteed nonexistent per rfc 6761
    url = "http://whatever.invalid./"
    site = brozzler.Site(None, {"seed": url})
    assert brozzler.is_permitted_by_robots(site, url)


def test_robots_connection_failure():
    # .invalid. is guaranteed nonexistent per rfc 6761
    url = "http://localhost:4/"  # nobody listens on port 4
    site = brozzler.Site(None, {"seed": url})
    assert brozzler.is_permitted_by_robots(site, url)


def test_scoping():
    test_scope = yaml.safe_load(
        r"""
max_hops: 100
accepts:
- url_match: REGEX_MATCH
  value: ^.*/audio_file/.*\.mp3$
- url_match: SURT_MATCH
  value: http://(com,vimeocdn,
- url_match: STRING_MATCH
  value: ec-media.soundcloud.com
- regex: ^https?://twitter\.com.*$
- substring: facebook.com
- regex: ^https?://(www.)?youtube.com/watch?.*$
  parent_url_regex: ^https?://(www.)?youtube.com/user/.*$
blocks:
- domain: twitter.com
  url_match: REGEX_MATCH
  value: ^.*lang=(?!en).*$
"""
    )

    site = brozzler.Site(
        None,
        {
            "id": 1,
            "seed": "http://example.com/foo/bar?baz=quux#monkey",
            "scope": test_scope,
        },
    )
    page = brozzler.Page(
        None, {"url": "http://example.com/foo/bar?baz=quux#monkey", "site_id": site.id}
    )

    assert site.accept_reject_or_neither("http://example.com/foo/bar", page) is True
    assert site.accept_reject_or_neither("http://example.com/foo/baz", page) is None

    assert site.accept_reject_or_neither("http://foo.com/some.mp3", page) is None
    assert (
        site.accept_reject_or_neither("http://foo.com/blah/audio_file/some.mp3", page)
        is True
    )

    assert (
        site.accept_reject_or_neither("http://a.b.vimeocdn.com/blahblah", page) is True
    )
    assert (
        site.accept_reject_or_neither("https://a.b.vimeocdn.com/blahblah", page) is None
    )

    assert site.accept_reject_or_neither("https://twitter.com/twit", page) is True
    assert (
        site.accept_reject_or_neither("https://twitter.com/twit?lang=en", page) is True
    )
    assert (
        site.accept_reject_or_neither("https://twitter.com/twit?lang=es", page) is False
    )

    assert (
        site.accept_reject_or_neither("https://www.facebook.com/whatevz", page) is True
    )

    assert (
        site.accept_reject_or_neither(
            "https://www.youtube.com/watch?v=dUIn5OAPS5s", page
        )
        is None
    )
    yt_user_page = brozzler.Page(
        None,
        {
            "url": "https://www.youtube.com/user/SonoraSantaneraVEVO",
            "site_id": site.id,
            "hops_from_seed": 10,
        },
    )
    assert (
        site.accept_reject_or_neither(
            "https://www.youtube.com/watch?v=dUIn5OAPS5s", yt_user_page
        )
        is True
    )


# Some changes to the brozzler ydl interface not represented in this test
# https://github.com/internetarchive/brozzler/issues/330
@pytest.mark.xfail
def test_ydl_proxy_down():
    sock = socket.socket()
    sock.bind(("127.0.0.1", 0))
    for not_listening_proxy in ("127.0.0.1:4", "127.0.0.1:%s" % sock.getsockname()[1]):
        worker = brozzler.BrozzlerWorker(frontier=None, proxy=not_listening_proxy)
        site = brozzler.Site(
            None, {"id": str(uuid.uuid4()), "seed": "http://example.com/"}
        )
        page = brozzler.Page(None, {"url": "http://example.com/"})

        # youtube-dl fetch
        with tempfile.TemporaryDirectory(prefix="brzl-ydl-"):
            with pytest.raises(brozzler.ProxyError):
                brozzler.ydl.do_youtube_dl(worker, site, page)


def test_proxy_down():
    """
    Test all fetching scenarios raise `brozzler.ProxyError` when proxy is down.

    This test needs to cover every possible fetch through the proxy other than
    fetches from the browser. For that, see test_brozzling.py.

    Tests two different kinds of connection error:
    - nothing listening the port (nobody listens on on port 4 :))
    - port bound but not accepting connections
    """
    sock = socket.socket()
    sock.bind(("127.0.0.1", 0))
    for not_listening_proxy in ("127.0.0.1:4", "127.0.0.1:%s" % sock.getsockname()[1]):
        worker = brozzler.BrozzlerWorker(frontier=None, proxy=not_listening_proxy)
        site = brozzler.Site(
            None, {"id": str(uuid.uuid4()), "seed": "http://example.com/"}
        )
        page = brozzler.Page(None, {"url": "http://example.com/"})

        # robots.txt fetch
        with pytest.raises(brozzler.ProxyError):
            brozzler.is_permitted_by_robots(
                site, "http://example.com/", proxy=not_listening_proxy
            )

        # raw fetch
        with pytest.raises(brozzler.ProxyError):
            worker._fetch_url(site, page=page)

        # WARCPROX_WRITE_RECORD
        with pytest.raises(brozzler.ProxyError):
            worker._warcprox_write_record(
                warcprox_address=not_listening_proxy,
                url="test://proxy_down/warcprox_write_record",
                warc_type="metadata",
                content_type="text/plain",
                payload=b"""payload doesn't matter here""",
            )


def test_start_stop_backwards_compat():
    site = brozzler.Site(None, {"seed": "http://example.com/"})
    assert len(site.starts_and_stops) == 1
    assert site.starts_and_stops[0]["start"]
    assert site.starts_and_stops[0]["stop"] is None
    assert "start_time" not in site

    site = brozzler.Site(
        None,
        {"seed": "http://example.com/", "start_time": datetime.datetime(2017, 1, 1)},
    )
    assert len(site.starts_and_stops) == 1
    assert site.starts_and_stops[0]["start"] == datetime.datetime(2017, 1, 1)
    assert site.starts_and_stops[0]["stop"] is None
    assert "start_time" not in site

    job = brozzler.Job(None, {"seeds": [{"url": "https://example.com/"}]})
    assert job.starts_and_stops[0]["start"]
    assert job.starts_and_stops[0]["stop"] is None
    assert "started" not in job
    assert "finished" not in job

    job = brozzler.Job(
        None,
        {
            "seeds": [{"url": "https://example.com/"}],
            "started": datetime.datetime(2017, 1, 1),
            "finished": datetime.datetime(2017, 1, 2),
        },
    )
    assert job.starts_and_stops[0]["start"] == datetime.datetime(2017, 1, 1)
    assert job.starts_and_stops[0]["stop"] == datetime.datetime(2017, 1, 2)
    assert "started" not in job
    assert "finished" not in job


class Exception1(Exception):
    pass


class Exception2(Exception):
    pass


def test_thread_raise_not_accept():
    def never_accept():
        try:
            brozzler.sleep(2)
        except Exception as e:
            nonlocal thread_caught_exception
            thread_caught_exception = e

    # test that thread_raise does not raise exception in a thread that has no
    # `with thread_exception_gate()` block
    thread_caught_exception = None
    th = threading.Thread(target=never_accept)
    th.start()
    brozzler.thread_raise(th, Exception1)
    th.join()
    assert thread_caught_exception is None


def test_thread_raise_immediate():
    def accept_immediately():
        try:
            with brozzler.thread_accept_exceptions():
                brozzler.sleep(2)
        except Exception as e:
            nonlocal thread_caught_exception
            thread_caught_exception = e

    # test immediate exception raise
    thread_caught_exception = None
    th = threading.Thread(target=accept_immediately)
    th.start()
    brozzler.thread_raise(th, Exception1)
    start = time.time()
    th.join()
    assert thread_caught_exception
    assert isinstance(thread_caught_exception, Exception1)
    assert time.time() - start < 1.0


def test_thread_raise_safe_exit():
    def delay_context_exit():
        gate = brozzler.thread_accept_exceptions()
        orig_exit = type(gate).__exit__
        try:
            type(gate).__exit__ = lambda self, et, ev, t: (
                brozzler.sleep(2),
                orig_exit(self, et, ev, t),
                False,
            )[-1]
            with brozzler.thread_accept_exceptions() as gate:
                brozzler.sleep(2)
        except Exception as e:
            nonlocal thread_caught_exception
            thread_caught_exception = e
        finally:
            type(gate).__exit__ = orig_exit

    # test that a second thread_raise() doesn't result in an exception in
    # ThreadExceptionGate.__exit__
    thread_caught_exception = None
    th = threading.Thread(target=delay_context_exit)
    th.start()
    time.sleep(0.2)
    brozzler.thread_raise(th, Exception1)
    time.sleep(0.2)
    brozzler.thread_raise(th, Exception2)
    th.join()
    assert thread_caught_exception
    assert isinstance(thread_caught_exception, Exception1)


def test_thread_raise_pending_exception():
    def accept_eventually():
        try:
            brozzler.sleep(2)
            with brozzler.thread_accept_exceptions():
                pass
        except Exception as e:
            nonlocal thread_caught_exception
            thread_caught_exception = e

    # test exception that has to wait for `with thread_exception_gate()` block
    thread_caught_exception = None
    th = threading.Thread(target=accept_eventually)
    th.start()
    brozzler.thread_raise(th, Exception1)
    start = time.time()
    th.join()
    assert isinstance(thread_caught_exception, Exception1)
    assert time.time() - start > 1.0


def test_thread_raise_second_with_block():
    def two_with_blocks():
        try:
            with brozzler.thread_accept_exceptions():
                time.sleep(2)
            return  # test fails
        except Exception1:
            pass
        except:  # noqa: E722
            return  # fail test

        try:
            with brozzler.thread_accept_exceptions():
                brozzler.sleep(2)
        except Exception as e:
            nonlocal thread_caught_exception
            thread_caught_exception = e

    # test that second `with` block gets second exception raised during first
    # `with` block
    thread_caught_exception = None
    th = threading.Thread(target=two_with_blocks)
    th.start()
    brozzler.thread_raise(th, Exception1)
    brozzler.thread_raise(th, Exception2)
    th.join()
    assert isinstance(thread_caught_exception, Exception2)


# brozzler.ydl.YoutubeDLSpy is missing
# https://github.com/internetarchive/brozzler/issues/330
@pytest.mark.xfail
def test_needs_browsing():
    # only one test case here right now, which exposed a bug

    class ConvenientHeaders(http.client.HTTPMessage):
        def __init__(self, headers):
            http.client.HTTPMessage.__init__(self)
            for k, v in headers.items():
                self.add_header(k, v)

    page = brozzler.Page(None, {"url": "http://example.com/a"})

    spy = brozzler.ydl.YoutubeDLSpy()
    spy.fetches.append(
        {
            "url": "http://example.com/a",
            "method": "HEAD",
            "response_code": 301,
            "response_headers": ConvenientHeaders({"Location": "/b"}),
        }
    )
    spy.fetches.append(
        {
            "url": "http://example.com/b",
            "method": "GET",
            "response_code": 200,
            "response_headers": ConvenientHeaders({"Content-Type": "application/pdf"}),
        }
    )

    assert not brozzler.worker.BrozzlerWorker._needs_browsing(None, page, spy.fetches)


def test_seed_redirect():
    site = brozzler.Site(None, {"seed": "http://foo.com/"})
    site.note_seed_redirect("https://foo.com/a/b/c")
    assert site.scope == {
        "accepts": [
            {
                "ssurt": "com,foo,//http:/",
            },
            {
                "ssurt": "com,foo,//https:/",
            },
        ]
    }

    site = brozzler.Site(None, {"seed": "https://foo.com/"})
    site.note_seed_redirect("http://foo.com/a/b/c")
    assert site.scope == {
        "accepts": [
            {
                "ssurt": "com,foo,//https:/",
            },
            {
                "ssurt": "com,foo,//http:/",
            },
        ]
    }

    site = brozzler.Site(None, {"seed": "http://foo.com/"})
    site.note_seed_redirect("https://bar.com/a/b/c")
    assert site.scope == {
        "accepts": [
            {
                "ssurt": "com,foo,//http:/",
            },
            {
                "ssurt": "com,bar,//https:/a/b/c",
            },
        ]
    }


def test_limit_failures():
    page = mock.Mock()
    page.failed_attempts = None
    page.brozzle_count = 0

    site = mock.Mock()
    site.status = "ACTIVE"
    site.active_brozzling_time = 0
    site.starts_and_stops = [{"start": datetime.datetime.now(datetime.timezone.utc)}]

    rr = mock.Mock()
    rr.servers = [mock.Mock()]
    rethink_query = mock.Mock(run=mock.Mock(return_value=[]))
    rr.db_list = mock.Mock(return_value=rethink_query)
    rr.table_list = mock.Mock(return_value=rethink_query)
    rr.table = mock.Mock(
        return_value=mock.Mock(
            between=mock.Mock(
                return_value=mock.Mock(limit=mock.Mock(return_value=rethink_query))
            )
        )
    )
    assert rr.table().between().limit().run() == []
    frontier = brozzler.RethinkDbFrontier(rr)
    frontier.enforce_time_limit = mock.Mock()
    frontier.honor_stop_request = mock.Mock()
    frontier.claim_page = mock.Mock(return_value=page)
    frontier._maybe_finish_job = mock.Mock()

    browser = mock.Mock()

    worker = brozzler.BrozzlerWorker(frontier)
    worker.brozzle_page = mock.Mock(side_effect=Exception)

    assert page.failed_attempts is None
    assert page.brozzle_count == 0
    assert site.status == "ACTIVE"

    worker.brozzle_site(browser, site)
    assert page.failed_attempts == 1
    assert page.brozzle_count == 0
    assert site.status == "ACTIVE"

    worker.brozzle_site(browser, site)
    assert page.failed_attempts == 2
    assert page.brozzle_count == 0
    assert site.status == "ACTIVE"

    worker.brozzle_site(browser, site)
    assert page.failed_attempts == 3
    assert page.brozzle_count == 1
    assert site.status == "FINISHED"



#!/usr/bin/env python
"""
test_cluster.py - integration tests for a brozzler cluster, expects brozzler,
warcprox, pywb, rethinkdb and other dependencies to be running already

Copyright (C) 2016-2018 Internet Archive

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import datetime
import http.server
import os
import socket
import subprocess
import threading
import time
import urllib.request

import doublethink
import pytest
import requests
import structlog
import warcprox

import brozzler.cli

logger = structlog.get_logger(logger_name=__name__)


# https://stackoverflow.com/questions/166506/finding-local-ip-addresses-using-pythons-stdlib
def _local_address():
    import socket

    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(("10.255.255.255", 1))  # ip doesn't need to be reachable
        return s.getsockname()[0]
    except:  # noqa: E722
        return "127.0.0.1"
    finally:
        s.close()


local_address = _local_address()


def start_service(service):
    subprocess.check_call(["sudo", "svc", "-u", "/etc/service/" + service])


def stop_service(service):
    subprocess.check_call(["sudo", "svc", "-d", "/etc/service/" + service])
    while True:
        status = subprocess.check_output(["sudo", "svstat", "/etc/service/" + service])
        if b" down " in status:
            break
        time.sleep(0.5)


@pytest.fixture(scope="module")
def rethinker(request):
    db = request.param if hasattr(request, "param") else "ignoreme"
    servers = os.environ.get("BROZZLER_RETHINKDB_SERVERS", "localhost")
    return doublethink.Rethinker(db=db, servers=servers.split(","))


@pytest.fixture(scope="module")
def httpd(request):
    class RequestHandler(http.server.SimpleHTTPRequestHandler):
        def do_POST(self):
            logger.info(
                "RequestHandler.do_POST",
                requestline=self.requestline,
                headers=self.headers,
            )
            self.do_GET()

        def do_GET(self):
            logger.info(
                "RequestHandler.do_GET",
                requestline=self.requestline,
                headers=self.headers,
            )
            if self.path == "/site5/redirect/":
                self.send_response(303, "See other")
                self.send_header("Connection", "close")
                self.send_header("Content-Length", 0)
                self.send_header("Location", "/site5/destination/")
                self.end_headers()
                self.wfile.write(b"")
            elif self.path == "/site9/redirect.html":
                self.send_response(303, "See other")
                self.send_header("Connection", "close")
                self.send_header("Content-Length", 0)
                self.send_header("Location", "/site9/destination.html")
                self.end_headers()
                self.wfile.write(b"")
            elif self.path.startswith("/infinite/"):
                payload = b"""
<html>
 <head>
  <title>infinite site</title>
 </head>
 <body>
  <a href='a/'>a/</a> <a href='b/'>b/</a> <a href='c/'>c/</a>
  <a href='d/'>d/</a> <a href='e/'>e/</a> <a href='f/'>f/</a>
  <a href='g/'>g/</a> <a href='h/'>h/</a> <a href='i/'>i/</a>
 </body>
</html>
"""
                self.send_response(200, "OK")
                self.send_header("Connection", "close")
                self.send_header("Content-Length", len(payload))
                self.end_headers()
                self.wfile.write(payload)
            else:
                super().do_GET()

    # SimpleHTTPRequestHandler always uses CWD so we have to chdir
    os.chdir(os.path.join(os.path.dirname(__file__), "htdocs"))

    httpd = http.server.HTTPServer((local_address, 0), RequestHandler)
    httpd_thread = threading.Thread(name="httpd", target=httpd.serve_forever)
    httpd_thread.start()

    def fin():
        httpd.shutdown()
        httpd.server_close()
        httpd_thread.join()

    request.addfinalizer(fin)

    return httpd


def make_url(httpd, rel_url):
    return "http://%s:%s%s" % (local_address, httpd.server_port, rel_url)


def test_httpd(httpd):
    """
    Tests that our http server is working as expected, and that two fetches
    of the same url return the same payload, proving it can be used to test
    deduplication.
    """
    payload1 = None
    url = make_url(httpd, "/site1/file1.txt")
    with urllib.request.urlopen(url) as response:
        assert response.status == 200
        payload1 = response.read()
        assert payload1

    with urllib.request.urlopen(url) as response:
        assert response.status == 200
        payload2 = response.read()
        assert payload2

    assert payload1 == payload2


@pytest.mark.skip()
def test_services_up(rethinker):
    """Check that the expected services are up and running."""
    # check that rethinkdb is listening and looks sane
    rr = rethinker
    tbls = rr.table_list().run()
    assert len(tbls) > 10

    # check that warcprox is listening
    with socket.socket() as s:
        # if the connect fails an exception is raised and the test fails
        s.connect(("localhost", 8000))

    # check that pywb is listening
    with socket.socket() as s:
        # if the connect fails an exception is raised and the test fails
        s.connect(("localhost", 8880))

    # check that brozzler dashboard is listening
    with socket.socket() as s:
        # if the connect fails an exception is raised and the test fails
        s.connect(("localhost", 8881))


@pytest.mark.parametrize("rethinker", ["brozzler"], indirect=True)
@pytest.mark.skip(reason="expects brozzler worker daemon running")
def test_brozzle_site(httpd, rethinker):
    test_id = "test_brozzle_site-%s" % datetime.datetime.utcnow().isoformat()
    rr = rethinker
    site = brozzler.Site(
        rr,
        {
            "seed": make_url(httpd, "/site1/"),
            "warcprox_meta": {"captures-table-extra-fields": {"test_id": test_id}},
        },
    )

    # the two pages we expect to be crawled
    page1 = make_url(httpd, "/site1/")
    page2 = make_url(httpd, "/site1/file1.txt")
    robots = make_url(httpd, "/robots.txt")

    # so we can examine rethinkdb before it does anything
    try:
        stop_service("brozzler-worker")

        assert site.id is None
        frontier = brozzler.RethinkDbFrontier(rr)
        brozzler.new_site(frontier, site)
        assert site.id is not None
        assert len(list(frontier.site_pages(site.id))) == 1
    finally:
        start_service("brozzler-worker")

    # the site should be brozzled fairly quickly
    start = time.time()
    while site.status != "FINISHED" and time.time() - start < 300:
        time.sleep(0.5)
        site.refresh()
    assert site.status == "FINISHED"

    # check that we got the two pages we expected
    pages = list(frontier.site_pages(site.id))
    assert len(pages) == 2
    assert {page.url for page in pages} == {
        make_url(httpd, "/site1/"),
        make_url(httpd, "/site1/file1.txt"),
    }

    time.sleep(2)  # in case warcprox hasn't finished processing urls
    # take a look at the captures table
    captures = rr.table("captures").filter({"test_id": test_id}).run()
    captures_by_url = {c["url"]: c for c in captures if c["http_method"] != "HEAD"}
    assert robots in captures_by_url
    assert page1 in captures_by_url
    assert page2 in captures_by_url
    assert "screenshot:%s" % page1 in captures_by_url
    assert "thumbnail:%s" % page1 in captures_by_url
    # no screenshots of plaintext

    # check pywb
    t14 = captures_by_url[page2]["timestamp"].strftime("%Y%m%d%H%M%S")
    wb_url = "http://localhost:8880/brozzler/%s/%s" % (t14, page2)
    expected_payload = open(
        os.path.join(os.path.dirname(__file__), "htdocs", "site1", "file1.txt"), "rb"
    ).read()
    assert requests.get(wb_url).content == expected_payload

    url = "screenshot:%s" % page1
    t14 = captures_by_url[url]["timestamp"].strftime("%Y%m%d%H%M%S")
    wb_url = "http://localhost:8880/brozzler/%s/%s" % (t14, url)
    response = requests.get(wb_url)
    assert response.status_code == 200
    assert response.headers["content-type"] == "image/jpeg"

    url = "thumbnail:%s" % page1
    t14 = captures_by_url[url]["timestamp"].strftime("%Y%m%d%H%M%S")
    wb_url = "http://localhost:8880/brozzler/%s/%s" % (t14, url)
    response = requests.get(wb_url)
    assert response.status_code == 200
    assert response.headers["content-type"] == "image/jpeg"


@pytest.mark.skip(reason="expects warcprox daemon running")
def test_proxy_warcprox(httpd):
    """Test --proxy with proxy that happens to be warcprox"""
    try:
        stop_service("brozzler-worker")
        _test_proxy_setting(
            httpd, proxy="localhost:8000", warcprox_auto=False, is_warcprox=True
        )
    finally:
        start_service("brozzler-worker")


@pytest.mark.skip(reason="expects warcprox daemon running")
def test_proxy_non_warcprox(httpd):
    """Test --proxy with proxy that happens not to be warcprox"""

    class DumbProxyRequestHandler(http.server.SimpleHTTPRequestHandler):
        def do_HEAD(self):
            if not hasattr(self.server, "requests"):
                self.server.requests = []
            logger.info("%s %s", self.command, self.path)
            self.server.requests.append("%s %s" % (self.command, self.path))
            response = urllib.request.urlopen(self.path)
            self.wfile.write(
                ("HTTP/1.0 %s %s\r\n" % (response.code, response.reason)).encode(
                    "ascii"
                )
            )
            for header in response.getheaders():
                self.wfile.write(
                    ("%s: %s\r\n" % (header[0], header[1])).encode("ascii")
                )
            self.wfile.write(b"\r\n")
            return response

        def do_GET(self):
            response = self.do_HEAD()
            self.copyfile(response, self.wfile)

        def do_WARCPROX_WRITE_RECORD(self):
            if not hasattr(self.server, "requests"):
                self.server.requests = []
            logger.info("%s %s", self.command, self.path)
            self.send_error(400)

    proxy = http.server.HTTPServer(("localhost", 0), DumbProxyRequestHandler)
    th = threading.Thread(name="dumb-proxy", target=proxy.serve_forever)
    th.start()

    try:
        stop_service("brozzler-worker")
        _test_proxy_setting(
            httpd,
            proxy="localhost:%s" % proxy.server_port,
            warcprox_auto=False,
            is_warcprox=False,
        )
    finally:
        start_service("brozzler-worker")
    assert len(proxy.requests) <= 15
    assert proxy.requests.count("GET /status") == 1
    assert ("GET %s" % make_url(httpd, "/site1/")) in proxy.requests
    assert ("GET %s" % make_url(httpd, "/site1/file1.txt")) in proxy.requests
    assert [
        req for req in proxy.requests if req.startswith("WARCPROX_WRITE_RECORD")
    ] == []

    proxy.shutdown()
    th.join()


@pytest.mark.skip()
def test_no_proxy(httpd):
    try:
        stop_service("brozzler-worker")
        _test_proxy_setting(httpd, proxy=None, warcprox_auto=False, is_warcprox=False)
    finally:
        start_service("brozzler-worker")
    # XXX how to check that no proxy was used?


@pytest.mark.skip()
def test_warcprox_auto(httpd):
    """Test --warcprox-auto"""
    try:
        stop_service("brozzler-worker")
        _test_proxy_setting(httpd, proxy=None, warcprox_auto=True, is_warcprox=True)
    finally:
        start_service("brozzler-worker")


@pytest.mark.skip()
def test_proxy_conflict():
    with pytest.raises(AssertionError):
        brozzler.worker.BrozzlerWorker(
            None, None, warcprox_auto=True, proxy="localhost:12345"
        )


@pytest.mark.skip()
@pytest.mark.parametrize("rethinker", ["brozzler"], indirect=True)
def _test_proxy_setting(
    httpd, rethinker, proxy=None, warcprox_auto=False, is_warcprox=False
):
    test_id = "test_proxy=%s_warcprox_auto=%s_is_warcprox=%s-%s" % (
        proxy,
        warcprox_auto,
        is_warcprox,
        datetime.datetime.utcnow().isoformat(),
    )

    # the two pages we expect to be crawled
    page1 = make_url(httpd, "/site1/")
    page2 = make_url(httpd, "/site1/file1.txt")
    robots = make_url(httpd, "/robots.txt")

    rr = rethinker
    service_registry = doublethink.ServiceRegistry(rr)
    site = brozzler.Site(
        rr,
        {
            "seed": make_url(httpd, "/site1/"),
            "warcprox_meta": {"captures-table-extra-fields": {"test_id": test_id}},
        },
    )
    assert site.id is None
    frontier = brozzler.RethinkDbFrontier(rr)
    brozzler.new_site(frontier, site)
    assert site.id is not None
    assert len(list(frontier.site_pages(site.id))) == 1

    worker = brozzler.worker.BrozzlerWorker(
        frontier,
        service_registry,
        max_browsers=1,
        chrome_exe=brozzler.suggest_default_chrome_exe(),
        warcprox_auto=warcprox_auto,
        proxy=proxy,
    )
    browser = worker._browser_pool.acquire()
    worker.brozzle_site(browser, site)
    worker._browser_pool.release(browser)

    # check proxy is set
    assert site.status == "FINISHED"
    if warcprox_auto:
        assert site.proxy[-5:] == ":8000"
    else:
        assert not site.proxy
    site.refresh()  # check that these things were persisted
    assert site.status == "FINISHED"
    if warcprox_auto:
        assert site.proxy[-5:] == ":8000"
    else:
        assert not site.proxy

    # check that we got the two pages we expected
    pages = list(frontier.site_pages(site.id))
    assert len(pages) == 2
    assert {page.url for page in pages} == {
        make_url(httpd, "/site1/"),
        make_url(httpd, "/site1/file1.txt"),
    }

    time.sleep(2)  # in case warcprox hasn't finished processing urls
    # take a look at the captures table
    captures = rr.table("captures").filter({"test_id": test_id}).run()
    captures_by_url = {c["url"]: c for c in captures if c["http_method"] != "HEAD"}
    if is_warcprox:
        assert robots in captures_by_url
        assert page1 in captures_by_url
        assert page2 in captures_by_url
        assert "screenshot:%s" % page1 in captures_by_url
        assert "thumbnail:%s" % page1 in captures_by_url

        # check pywb
        t14 = captures_by_url[page2]["timestamp"].strftime("%Y%m%d%H%M%S")
        wb_url = "http://localhost:8880/brozzler/%s/%s" % (t14, page2)
        expected_payload = open(
            os.path.join(os.path.dirname(__file__), "htdocs", "site1", "file1.txt"),
            "rb",
        ).read()
        assert requests.get(wb_url).content == expected_payload
    else:
        assert captures_by_url == {}


@pytest.mark.parametrize("rethinker", ["brozzler"], indirect=True)
@pytest.mark.skip(reason="expects brozzler worker daemon running")
def test_obey_robots(httpd, rethinker):
    test_id = "test_obey_robots-%s" % datetime.datetime.utcnow().isoformat()
    rr = rethinker
    site = brozzler.Site(
        rr,
        {
            "seed": make_url(httpd, "/site1/"),
            "user_agent": "im a badbot",  # robots.txt blocks badbot
            "warcprox_meta": {"captures-table-extra-fields": {"test_id": test_id}},
        },
    )

    # so we can examine rethinkdb before it does anything
    try:
        stop_service("brozzler-worker")

        assert site.id is None
        frontier = brozzler.RethinkDbFrontier(rr)
        brozzler.new_site(frontier, site)
        assert site.id is not None
        site_pages = list(frontier.site_pages(site.id))
        assert len(site_pages) == 1
        assert site_pages[0].url == site.seed
        assert site_pages[0].needs_robots_check
    finally:
        start_service("brozzler-worker")

    # the site should be brozzled fairly quickly
    start = time.time()
    while site.status != "FINISHED" and time.time() - start < 300:
        time.sleep(0.5)
        site.refresh()
    assert site.status == "FINISHED"

    # check that only the one page is in rethinkdb
    pages = list(frontier.site_pages(site.id))
    assert len(pages) == 1
    page = pages[0]
    assert page.url == make_url(httpd, "/site1/")
    assert page.blocked_by_robots

    # take a look at the captures table
    time.sleep(2)  # in case warcprox hasn't finished processing urls
    robots_url = make_url(httpd, "/robots.txt")
    captures = list(rr.table("captures").filter({"test_id": test_id}).run())
    assert len(captures) == 1
    assert captures[0]["url"] == robots_url

    # check pywb
    t14 = captures[0]["timestamp"].strftime("%Y%m%d%H%M%S")
    wb_url = "http://localhost:8880/brozzler/%s/%s" % (t14, robots_url)
    expected_payload = open(
        os.path.join(os.path.dirname(__file__), "htdocs", "robots.txt"), "rb"
    ).read()
    assert requests.get(wb_url, allow_redirects=False).content == expected_payload


@pytest.mark.parametrize("rethinker", ["brozzler"], indirect=True)
@pytest.mark.skip(reason="expects brozzler worker daemon running")
def test_login(httpd, rethinker):
    test_id = "test_login-%s" % datetime.datetime.utcnow().isoformat()
    rr = rethinker
    site = brozzler.Site(
        rr,
        {
            "seed": make_url(httpd, "/site2/"),
            "warcprox_meta": {"captures-table-extra-fields": {"test_id": test_id}},
            "username": "test_username",
            "password": "test_password",
        },
    )

    frontier = brozzler.RethinkDbFrontier(rr)
    brozzler.new_site(frontier, site)

    # the site should be brozzled fairly quickly
    start = time.time()
    while site.status != "FINISHED" and time.time() - start < 300:
        time.sleep(0.5)
        site.refresh()
    assert site.status == "FINISHED"

    # take a look at the captures table
    time.sleep(2)  # in case warcprox hasn't finished processing urls
    captures = list(
        rr.table("captures").filter({"test_id": test_id}).order_by("timestamp").run()
    )
    meth_url = ["%s %s" % (c["http_method"], c["url"]) for c in captures]

    # there are several forms in in htdocs/site2/login.html but only one
    # that brozzler's heuristic should match and try to submit, and it has
    # action='00', so we can check for that here
    assert ("POST %s" % make_url(httpd, "/site2/00")) in meth_url

    # sanity check the rest of the crawl
    assert ("GET %s" % make_url(httpd, "/robots.txt")) in meth_url
    assert ("GET %s" % make_url(httpd, "/site2/")) in meth_url
    assert (
        "WARCPROX_WRITE_RECORD screenshot:%s" % make_url(httpd, "/site2/")
    ) in meth_url
    assert (
        "WARCPROX_WRITE_RECORD thumbnail:%s" % make_url(httpd, "/site2/")
    ) in meth_url
    assert ("GET %s" % make_url(httpd, "/site2/login.html")) in meth_url
    assert (
        "WARCPROX_WRITE_RECORD screenshot:%s" % make_url(httpd, "/site2/login.html")
    ) in meth_url
    assert (
        "WARCPROX_WRITE_RECORD thumbnail:%s" % make_url(httpd, "/site2/login.html")
    ) in meth_url


@pytest.mark.parametrize("rethinker", ["brozzler"], indirect=True)
@pytest.mark.skip(reason="expects brozzler worker daemon running")
def test_seed_redirect(httpd, rethinker):
    test_id = "test_seed_redirect-%s" % datetime.datetime.utcnow().isoformat()
    rr = rethinker
    seed_url = make_url(httpd, "/site5/redirect/")
    site = brozzler.Site(
        rr,
        {
            "seed": make_url(httpd, "/site5/redirect/"),
            "warcprox_meta": {"captures-table-extra-fields": {"test_id": test_id}},
        },
    )
    assert site.scope == {
        "accepts": [
            {
                "ssurt": "%s//%s:http:/site5/redirect/"
                % (local_address, httpd.server_port)
            }
        ]
    }

    frontier = brozzler.RethinkDbFrontier(rr)
    brozzler.new_site(frontier, site)
    assert site.id

    # the site should be brozzled fairly quickly
    start = time.time()
    while site.status != "FINISHED" and time.time() - start < 300:
        time.sleep(0.5)
        site.refresh()
    assert site.status == "FINISHED"

    # take a look at the pages table
    pages = list(frontier.site_pages(site.id))
    assert len(pages) == 2
    pages.sort(key=lambda page: page.hops_from_seed)
    assert pages[0].hops_from_seed == 0
    assert pages[0].url == seed_url
    assert pages[0].redirect_url == make_url(httpd, "/site5/destination/")
    assert pages[1].hops_from_seed == 1
    assert pages[1].url == make_url(httpd, "/site5/destination/page2.html")

    # check that scope has been updated properly
    assert site.scope == {
        "accepts": [
            {
                "ssurt": "%s//%s:http:/site5/redirect/"
                % (local_address, httpd.server_port)
            },
            {
                "ssurt": "%s//%s:http:/site5/destination/"
                % (local_address, httpd.server_port)
            },
        ]
    }


@pytest.mark.parametrize("rethinker", ["brozzler"], indirect=True)
@pytest.mark.skip(reason="expects brozzler worker daemon running")
def test_hashtags(httpd, rethinker):
    test_id = "test_hashtags-%s" % datetime.datetime.utcnow().isoformat()
    rr = rethinker
    seed_url = make_url(httpd, "/site7/")
    site = brozzler.Site(
        rr,
        {
            "seed": seed_url,
            "warcprox_meta": {"captures-table-extra-fields": {"test_id": test_id}},
        },
    )

    frontier = brozzler.RethinkDbFrontier(rr)
    brozzler.new_site(frontier, site)
    assert site.id

    # the site should be brozzled fairly quickly
    start = time.time()
    while site.status != "FINISHED" and time.time() - start < 300:
        time.sleep(0.5)
        site.refresh()
    assert site.status == "FINISHED"

    # check that we the page we expected
    pages = sorted(list(frontier.site_pages(site.id)), key=lambda p: p.url)
    assert len(pages) == 2
    assert pages[0].url == seed_url
    assert pages[0].hops_from_seed == 0
    assert pages[0].brozzle_count == 1
    assert pages[0].outlinks["accepted"] == [make_url(httpd, "/site7/foo.html")]
    assert not pages[0].hashtags
    assert pages[1].url == make_url(httpd, "/site7/foo.html")
    assert pages[1].hops_from_seed == 1
    assert pages[1].brozzle_count == 1
    assert sorted(pages[1].hashtags) == [
        "#boosh",
        "#ignored",
        "#whee",
    ]

    time.sleep(2)  # in case warcprox hasn't finished processing urls
    # take a look at the captures table
    captures = rr.table("captures").filter({"test_id": test_id}).run()
    captures_by_url = {c["url"]: c for c in captures if c["http_method"] != "HEAD"}
    assert seed_url in captures_by_url
    assert make_url(httpd, "/site7/foo.html") in captures_by_url
    assert make_url(httpd, "/site7/whee.txt") in captures_by_url
    assert make_url(httpd, "/site7/boosh.txt") in captures_by_url
    assert "screenshot:%s" % seed_url in captures_by_url
    assert "thumbnail:%s" % seed_url in captures_by_url
    assert "screenshot:%s" % make_url(httpd, "/site7/foo.html") in captures_by_url
    assert "thumbnail:%s" % make_url(httpd, "/site7/foo.html") in captures_by_url


@pytest.mark.parametrize("rethinker", ["brozzler"], indirect=True)
@pytest.mark.skip(reason="expects brozzler worker daemon running")
def test_redirect_hashtags(httpd, rethinker):
    test_id = "test_hashtags-%s" % datetime.datetime.utcnow().isoformat()
    rr = rethinker
    seed_url = make_url(httpd, "/site9/")
    site = brozzler.Site(
        rr,
        {
            "seed": seed_url,
            "warcprox_meta": {"captures-table-extra-fields": {"test_id": test_id}},
        },
    )

    frontier = brozzler.RethinkDbFrontier(rr)
    brozzler.new_site(frontier, site)
    assert site.id

    # the site should be brozzled fairly quickly
    start = time.time()
    while site.status != "FINISHED" and time.time() - start < 300:
        time.sleep(0.5)
        site.refresh()
    assert site.status == "FINISHED"

    # check that we the page we expected
    pages = sorted(list(frontier.site_pages(site.id)), key=lambda p: p.url)
    assert len(pages) == 2
    assert pages[0].url == seed_url
    assert pages[0].hops_from_seed == 0
    assert pages[0].brozzle_count == 1
    assert pages[0].outlinks["accepted"] == [make_url(httpd, "/site9/redirect.html")]
    assert not pages[0].hashtags
    assert pages[1].url == make_url(httpd, "/site9/redirect.html")
    assert pages[1].hops_from_seed == 1
    assert pages[1].brozzle_count == 1
    assert sorted(pages[1].hashtags) == [
        "#hash1",
        "#hash2",
    ]

    time.sleep(2)  # in case warcprox hasn't finished processing urls
    # take a look at the captures table
    captures = rr.table("captures").filter({"test_id": test_id}).run()
    redirect_captures = [
        c
        for c in captures
        if c["url"] == make_url(httpd, "/site9/redirect.html")
        and c["http_method"] == "GET"
    ]
    assert len(redirect_captures) == 2  # youtube-dl + browser, no hashtags

    # === expected captures ===
    #  1. GET http://localhost:41243/favicon.ico
    #  2. GET http://localhost:41243/robots.txt
    #  3. GET http://localhost:41243/site9/
    #  4. GET http://localhost:41243/site9/
    #  5. GET http://localhost:41243/site9/destination.html
    #  6. GET http://localhost:41243/site9/destination.html
    #  7. GET http://localhost:41243/site9/redirect.html
    #  8. GET http://localhost:41243/site9/redirect.html
    #  9. HEAD http://localhost:41243/site9/
    # 10. HEAD http://localhost:41243/site9/redirect.html
    # 11. WARCPROX_WRITE_RECORD screenshot:http://localhost:41243/site9/
    # 12. WARCPROX_WRITE_RECORD screenshot:http://localhost:41243/site9/redirect.html
    # 13. WARCPROX_WRITE_RECORD thumbnail:http://localhost:41243/site9/
    # 14. WARCPROX_WRITE_RECORD thumbnail:http://localhost:41243/site9/redirect.html


@pytest.mark.parametrize("rethinker", ["brozzler"], indirect=True)
@pytest.mark.skip(reason="expects brozzler worker daemon running")
def test_stop_crawl(httpd, rethinker):
    rr = rethinker
    frontier = brozzler.RethinkDbFrontier(rr)

    # create a new job with three sites that could be crawled forever
    job_conf = {
        "seeds": [
            {"url": make_url(httpd, "/infinite/foo/")},
            {"url": make_url(httpd, "/infinite/bar/")},
            {"url": make_url(httpd, "/infinite/baz/")},
        ]
    }
    job = brozzler.new_job(frontier, job_conf)
    assert job.id

    sites = list(frontier.job_sites(job.id))
    assert not sites[0].stop_requested
    assert not sites[1].stop_requested

    # request crawl stop for one site using the command line entrypoint
    brozzler.cli.brozzler_stop_crawl(["brozzler-stop-crawl", "--site=%s" % sites[0].id])
    sites[0].refresh()
    assert sites[0].stop_requested

    # stop request should be honored quickly
    start = time.time()
    while not sites[0].status.startswith("FINISHED") and time.time() - start < 120:
        time.sleep(0.5)
        sites[0].refresh()
    assert sites[0].status == "FINISHED_STOP_REQUESTED"

    # but the other sites and the job as a whole should still be crawling
    sites[1].refresh()
    assert sites[1].status == "ACTIVE"
    sites[2].refresh()
    assert sites[2].status == "ACTIVE"
    job.refresh()
    assert job.status == "ACTIVE"

    # request crawl stop for the job using the command line entrypoint
    brozzler.cli.brozzler_stop_crawl(["brozzler-stop-crawl", "--job=%s" % job.id])
    job.refresh()
    assert job.stop_requested

    # stop request should be honored quickly
    start = time.time()
    while not job.status.startswith("FINISHED") and time.time() - start < 120:
        time.sleep(0.5)
        job.refresh()
    assert job.status == "FINISHED"

    # the other sites should also be FINISHED_STOP_REQUESTED
    sites[0].refresh()
    assert sites[0].status == "FINISHED_STOP_REQUESTED"
    sites[1].refresh()
    assert sites[1].status == "FINISHED_STOP_REQUESTED"
    sites[2].refresh()
    assert sites[2].status == "FINISHED_STOP_REQUESTED"


@pytest.mark.parametrize("rethinker", ["brozzler"], indirect=True)
@pytest.mark.skip(reason="expects brozzler worker daemon running")
def test_warcprox_outage_resiliency(httpd, rethinker):
    """
    Tests resiliency to warcprox outage.

    If no instances of warcprox are healthy when starting to crawl a site,
    brozzler-worker should sit there and wait until a healthy instance appears.

    If an instance goes down, sites assigned to that instance should bounce
    over to a healthy instance.

    If all instances of warcprox go down, brozzler-worker should sit and wait.
    """
    rr = rethinker
    frontier = brozzler.RethinkDbFrontier(rr)

    # run two instances of warcprox
    opts = warcprox.Options()
    opts.address = "0.0.0.0"
    opts.port = 0
    opts.rethinkdb_services_url = "rethinkdb://localhost/brozzler/services"

    warcprox1 = warcprox.controller.WarcproxController(opts)
    warcprox2 = warcprox.controller.WarcproxController(opts)
    warcprox1_thread = threading.Thread(
        target=warcprox1.run_until_shutdown, name="warcprox1"
    )
    warcprox2_thread = threading.Thread(
        target=warcprox2.run_until_shutdown, name="warcprox2"
    )

    # put together a site to crawl
    test_id = "test_warcprox_death-%s" % datetime.datetime.utcnow().isoformat()
    site = brozzler.Site(
        rr,
        {
            "seed": make_url(httpd, "/infinite/"),
            "warcprox_meta": {"captures-table-extra-fields": {"test_id": test_id}},
        },
    )

    try:
        # we manage warcprox instances ourselves, so stop the one running on
        # the system, if any
        try:
            stop_service("warcprox")
        except Exception:
            logger.warning("problem stopping warcprox service: %s", exc_info=True)

        # queue the site for brozzling
        brozzler.new_site(frontier, site)

        # check that nothing happens
        # XXX tail brozzler-worker.log or something?
        time.sleep(30)
        site.refresh()
        assert site.status == "ACTIVE"
        assert not site.proxy
        assert len(list(frontier.site_pages(site.id))) == 1

        # start one instance of warcprox
        warcprox1_thread.start()

        # check that it started using that instance
        start = time.time()
        while not site.proxy and time.time() - start < 30:
            time.sleep(0.5)
            site.refresh()
        assert site.proxy.endswith(":%s" % warcprox1.proxy.server_port)

        # check that the site accumulates pages in the frontier, confirming
        # that crawling is really happening
        start = time.time()
        while len(list(frontier.site_pages(site.id))) <= 1 and time.time() - start < 60:
            time.sleep(0.5)
            site.refresh()
        assert len(list(frontier.site_pages(site.id))) > 1

        # stop warcprox #1, start warcprox #2
        warcprox2_thread.start()
        warcprox1.stop.set()
        warcprox1_thread.join()

        # check that it switched over to warcprox #2
        start = time.time()
        while (
            not site.proxy
            or not site.proxy.endswith(":%s" % warcprox2.proxy.server_port)
        ) and time.time() - start < 30:
            time.sleep(0.5)
            site.refresh()
        assert site.proxy.endswith(":%s" % warcprox2.proxy.server_port)

        # stop warcprox #2
        warcprox2.stop.set()
        warcprox2_thread.join()

        page_count = len(list(frontier.site_pages(site.id)))
        assert page_count > 1

        # check that it is waiting for a warcprox to appear
        time.sleep(30)
        site.refresh()
        assert site.status == "ACTIVE"
        assert not site.proxy
        assert len(list(frontier.site_pages(site.id))) == page_count

        # stop crawling the site, else it can pollute subsequent test runs
        brozzler.cli.brozzler_stop_crawl(["brozzler-stop-crawl", "--site=%s" % site.id])
        site.refresh()
        assert site.stop_requested

        # stop request should be honored quickly
        start = time.time()
        while not site.status.startswith("FINISHED") and time.time() - start < 120:
            time.sleep(0.5)
            site.refresh()
        assert site.status == "FINISHED_STOP_REQUESTED"
    finally:
        warcprox1.stop.set()
        warcprox2.stop.set()
        warcprox1_thread.join()
        warcprox2_thread.join()
        start_service("warcprox")


@pytest.mark.parametrize("rethinker", ["brozzler"], indirect=True)
@pytest.mark.skip(reason="expects brozzler worker daemon running")
def test_time_limit(httpd, rethinker):
    rr = rethinker
    frontier = brozzler.RethinkDbFrontier(rr)

    # create a new job with one seed that could be crawled forever
    job_conf = {"seeds": [{"url": make_url(httpd, "/infinite/foo/"), "time_limit": 20}]}
    job = brozzler.new_job(frontier, job_conf)
    assert job.id

    sites = list(frontier.job_sites(job.id))
    assert len(sites) == 1

    # time limit should be enforced pretty soon
    start = time.time()
    while not sites[0].status.startswith("FINISHED") and time.time() - start < 120:
        time.sleep(0.5)
        sites[0].refresh()
    assert sites[0].status == "FINISHED_TIME_LIMIT"

    # all sites finished so job should be finished too
    start = time.time()
    job.refresh()
    while not job.status == "FINISHED" and time.time() - start < 10:
        time.sleep(0.5)
        job.refresh()
    assert job.status == "FINISHED"


@pytest.mark.parametrize("rethinker", ["brozzler"], indirect=True)
@pytest.mark.skip(reason="expects brozzler worker daemon running")
def test_ydl_stitching(httpd, rethinker):
    test_id = "test_ydl_stitching-%s" % datetime.datetime.utcnow().isoformat()
    rr = rethinker
    frontier = brozzler.RethinkDbFrontier(rr)
    site = brozzler.Site(
        rr,
        {
            "seed": make_url(httpd, "/site10/"),
            "warcprox_meta": {
                "warc-prefix": "test_ydl_stitching",
                "captures-table-extra-fields": {"test_id": test_id},
            },
        },
    )
    brozzler.new_site(frontier, site)

    # the site should be brozzled fairly quickly
    start = time.time()
    while site.status != "FINISHED" and time.time() - start < 300:
        time.sleep(0.5)
        site.refresh()
    assert site.status == "FINISHED"

    # check page.videos
    pages = list(frontier.site_pages(site.id))
    assert len(pages) == 1
    page = pages[0]
    assert len(page.videos) == 6
    stitched_url = "youtube-dl:00001:%s" % make_url(httpd, "/site10/")
    assert {
        "blame": "youtube-dl",
        "content-length": 267900,
        "content-type": "video/mp4",
        "response_code": 204,
        "url": stitched_url,
    } in page.videos

    time.sleep(2)  # in case warcprox hasn't finished processing urls
    # take a look at the captures table
    captures = list(rr.table("captures").filter({"test_id": test_id}).run())
    l = [c for c in captures if c["url"] == stitched_url]  # noqa: E741
    assert len(l) == 1
    c = l[0]
    assert c["filename"].startswith("test_ydl_stitching")
    assert c["content_type"] == "video/mp4"
    assert c["http_method"] == "WARCPROX_WRITE_RECORD"



"""
brozzler/ydl.py - youtube-dl / yt-dlp support for brozzler

Copyright (C) 2024-2025 Internet Archive

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import datetime
import json
import os
import random
import tempfile
import threading
import time
import urllib.request

import doublethink
import structlog
import urlcanon
import yt_dlp
from yt_dlp.utils import ExtractorError, match_filter_func

import brozzler

from . import metrics

thread_local = threading.local()


PROXY_ATTEMPTS = 4
YTDLP_WAIT = 10
YTDLP_MAX_REDIRECTS = 5


logger = structlog.get_logger(logger_name=__name__)


def isyoutubehost(url):
    # split 1 splits scheme from url, split 2 splits path from hostname, split 3 splits query string on hostname
    return "youtube.com" in url.split("//")[-1].split("/")[0].split("?")[0]


class ExtraHeaderAdder(urllib.request.BaseHandler):
    def __init__(self, extra_headers):
        self.extra_headers = extra_headers
        self.http_request = self._http_request
        self.https_request = self._http_request

    def _http_request(self, req):
        for h, v in self.extra_headers.items():
            if h.capitalize() not in req.headers:
                req.add_header(h, v)
        return req


def _build_youtube_dl(worker, destdir, site, page, ytdlp_proxy_endpoints):
    """
    Builds a yt-dlp `yt_dlp.YoutubeDL` for brozzling `site` with `worker`.

    The `YoutubeDL` instance does a few special brozzler-specific things:

    - periodically updates `site.last_claimed` in rethinkdb
    - pushes captured video to warcprox using a WARCPROX_WRITE_RECORD request
    - some logging

    Args:
        worker (brozzler.BrozzlerWorker): the calling brozzler worker
        destdir (str): where to save downloaded videos
        site (brozzler.Site): the site we are brozzling
        page (brozzler.Page): the page we are brozzling

    Returns:
        a yt-dlp `yt_dlp.YoutubeDL` instance
    """

    class _YoutubeDL(yt_dlp.YoutubeDL):
        logger = structlog.get_logger(__module__ + "." + __qualname__)

        def __init__(self, url, params=None, auto_init=True):
            super().__init__(params, auto_init)

            self.url = url
            self.logger = self.logger.bind(url=url)

        def process_ie_result(self, ie_result, download=True, extra_info=None):
            if extra_info is None:
                extra_info = {}
            result_type = ie_result.get("_type", "video")

            if result_type in ("url", "url_transparent"):
                if "extraction_depth" in extra_info:
                    self.logger.info(
                        "Following redirect",
                        redirect_url=ie_result["url"],
                        extraction_depth=extra_info["extraction_depth"],
                    )
                    extra_info["extraction_depth"] = 1 + extra_info.get(
                        "extraction_depth", 0
                    )
                else:
                    extra_info["extraction_depth"] = 0
                if extra_info["extraction_depth"] >= YTDLP_MAX_REDIRECTS:
                    raise ExtractorError(
                        f"Too many hops for URL: {ie_result['url']}",
                        expected=True,
                    )
            return super().process_ie_result(ie_result, download, extra_info)

        def add_default_extra_info(self, ie_result, ie, url):
            # hook in some logging
            super().add_default_extra_info(ie_result, ie, url)
            extract_context = self.logger.bind(extractor=ie.IE_NAME)
            if ie_result.get("_type") == "playlist":
                extract_context.info("found playlist")
                if ie.IE_NAME in {
                    "youtube:playlist",
                    "youtube:tab",
                    "soundcloud:user",
                    "instagram:user",
                }:
                    # At this point ie_result['entries'] is an iterator that
                    # will fetch more metadata from youtube to list all the
                    # videos. We unroll that iterator here partly because
                    # otherwise `process_ie_result()` will clobber it, and we
                    # use it later to extract the watch pages as outlinks.
                    try:
                        ie_result["entries_no_dl"] = list(ie_result["entries"])
                    except Exception:
                        extract_context.warning(
                            "failed to unroll entries ie_result['entries']?",
                            exc_info=True,
                        )
                        ie_result["entries_no_dl"] = []
                    ie_result["entries"] = []
                    self.logger.info(
                        "not downloading media files from this "
                        "playlist because we expect to capture them from "
                        "individual watch/track/detail pages",
                        media_file_count=len(ie_result["entries_no_dl"]),
                    )
            else:
                extract_context.info("found a download")

        def _push_video_to_warcprox(self, site, info_dict, postprocessor):
            # 220211 update: does yt-dlp supply content-type? no, not as such
            # XXX Don't know how to get the right content-type. Youtube-dl
            # doesn't supply it. Sometimes (with --hls-prefer-native)
            # youtube-dl produces a stitched-up video that /usr/bin/file fails
            # to identify (says "application/octet-stream"). `ffprobe` doesn't
            # give us a mimetype.
            if info_dict.get("ext") == "mp4":
                mimetype = "video/mp4"
            else:
                try:
                    import magic

                    mimetype = magic.from_file(info_dict["filepath"], mime=True)
                except ImportError:
                    mimetype = "video/%s" % info_dict["ext"]
                    self.logger.warning(
                        "guessing mimetype due to error",
                        mimetype=mimetype,
                        exc_info=True,
                    )

            # youtube watch page postprocessor is MoveFiles

            if postprocessor == "FixupM3u8" or postprocessor == "Merger":
                url = "youtube-dl:%05d:%s" % (
                    info_dict.get("playlist_index") or 1,
                    info_dict["webpage_url"],
                )
            else:
                url = info_dict.get("url", "")

            # skip urls containing .m3u8, to avoid duplicates handled by FixupM3u8
            if url == "" or ".m3u8" in url:
                return

            size = os.path.getsize(info_dict["filepath"])
            self.logger.info(
                "pushing video to warcprox",
                format=info_dict["format"],
                mimetype=mimetype,
                size=size,
                warcprox=worker._proxy_for(site),
            )
            with open(info_dict["filepath"], "rb") as f:
                # include content-length header to avoid chunked
                # transfer, which warcprox currently rejects
                extra_headers = dict(site.extra_headers())
                extra_headers["content-length"] = size
                request, response = worker._warcprox_write_record(
                    warcprox_address=worker._proxy_for(site),
                    url=url,
                    warc_type="resource",
                    content_type=mimetype,
                    payload=f,
                    extra_headers=extra_headers,
                )

            # consulted by _remember_videos()
            ydl.pushed_videos.append(
                {
                    "url": url,
                    "response_code": response.code,
                    "content-type": mimetype,
                    "content-length": size,
                }
            )

    def maybe_heartbeat_site_last_claimed(*args, **kwargs):
        # in case yt-dlp takes a long time, heartbeat site.last_claimed
        # to prevent another brozzler-worker from claiming the site
        try:
            if (
                site.rr
                and doublethink.utcnow() - site.last_claimed
                > datetime.timedelta(minutes=worker.SITE_SESSION_MINUTES)
            ):
                worker.logger.debug(
                    "heartbeating site.last_claimed to prevent another "
                    "brozzler-worker claiming this site",
                    id=site.id,
                )
                site.last_claimed = doublethink.utcnow()
                site.save()
        except:  # noqa: E722
            worker.logger.debug(
                "problem heartbeating site.last_claimed site",
                id=site.id,
                exc_info=True,
            )

    def ydl_postprocess_hook(d):
        if d["status"] == "finished":
            worker.logger.info(
                "[ydl_postprocess_hook] Finished postprocessing",
                postprocessor=d["postprocessor"],
            )
            is_youtube_host = isyoutubehost(d["info_dict"]["webpage_url"])

            metrics.brozzler_ydl_download_successes.labels(is_youtube_host).inc(1)
            if worker._using_warcprox(site):
                _YoutubeDL._push_video_to_warcprox(
                    _YoutubeDL, site, d["info_dict"], d["postprocessor"]
                )

    # default socket_timeout is 20 -- we hit it often when cluster is busy
    ydl_opts = {
        "outtmpl": "{}/ydl%(autonumber)s.out".format(destdir),
        "retries": 1,
        "nocheckcertificate": True,
        "noplaylist": True,
        "noprogress": True,
        "nopart": True,
        "no_color": True,
        "socket_timeout": 40,
        "progress_hooks": [maybe_heartbeat_site_last_claimed],
        "postprocessor_hooks": [ydl_postprocess_hook],
        # YouTube has started providing videos in upscaled "Super Resolution"
        # format, which we don't want to download. yt-dlp identifies these via
        # the -sr suffix on the format_id. We can filter these out from the
        # list without putting any other prerequisites on the results.
        # https://alexwlchan.net/til/2025/ignore-ai-scaled-videos/
        "format": "[format_id!*=-sr]",
        # https://github.com/yt-dlp/yt-dlp#format-selection
        # "By default, yt-dlp tries to download the best available quality..."
        # v.2023.07.06 https://www.reddit.com/r/youtubedl/wiki/h264/?rdt=63577
        # recommended: convert working cli to api call with
        # https://github.com/yt-dlp/yt-dlp/blob/master/devscripts/cli_to_api.py
        "format_sort": ["res:720", "vcodec:h264", "acodec:aac"],
        # skip live streams
        "match_filter": match_filter_func("!is_live"),
        "extractor_args": {"generic": {"impersonate": [""]}},
        # --cache-dir local or..
        # this looked like a problem with nsf-mounted homedir, maybe not a problem for brozzler on focal?
        "cache_dir": "/home/archiveit",
        "logger": logger,
        "verbose": False,
        "quiet": False,
        # recommended to avoid bot detection
        "sleep_interval": 7,
        "max_sleep_interval": 27,
    }

    ytdlp_url = page.redirect_url if page.redirect_url else page.url
    is_youtube_host = isyoutubehost(ytdlp_url)
    if is_youtube_host and ytdlp_proxy_endpoints:
        ydl_opts["proxy"] = random.choice(ytdlp_proxy_endpoints)
        # don't log proxy value secrets
        ytdlp_proxy_for_logs = (
            ydl_opts["proxy"].split("@")[1] if "@" in ydl_opts["proxy"] else "@@@"
        )
        logger.info("using yt-dlp proxy ...", proxy=ytdlp_proxy_for_logs)

    # skip warcprox proxying yt-dlp v.2023.07.06: youtube extractor using ranges
    # if worker._proxy_for(site):
    #    ydl_opts["proxy"] = "http://{}".format(worker._proxy_for(site))

    ydl = _YoutubeDL(ytdlp_url, params=ydl_opts)
    if site.extra_headers():
        ydl._opener.add_handler(ExtraHeaderAdder(site.extra_headers(page)))
    ydl.pushed_videos = []
    ydl.is_youtube_host = is_youtube_host

    return ydl


def _remember_videos(page, pushed_videos=None):
    """
    Saves info about videos captured by yt-dlp in `page.videos`.
    """
    if "videos" not in page:
        page.videos = []
    for pushed_video in pushed_videos or []:
        video = {
            "blame": "youtube-dl",
            "url": pushed_video["url"],
            "response_code": pushed_video["response_code"],
            "content-type": pushed_video["content-type"],
            "content-length": pushed_video["content-length"],
        }
        logger.debug("embedded video", video=video)
        page.videos.append(video)


def _try_youtube_dl(worker, ydl, site, page):
    max_attempts = PROXY_ATTEMPTS if ydl.is_youtube_host else 1
    attempt = 0
    while attempt < max_attempts:
        try:
            logger.info("trying yt-dlp", url=ydl.url)
            # should_download_vid = not ydl.is_youtube_host
            # then
            # ydl.extract_info(str(urlcanon.whatwg(ydl.url)), download=should_download_vid)
            # if ydl.is_youtube_host and ie_result:
            #     download_url = ie_result.get("url")
            with brozzler.thread_accept_exceptions():
                # we do whatwg canonicalization here to avoid "<urlopen error
                # no host given>" resulting in ProxyError
                # needs automated test
                # and yt-dlp needs sanitize_info for extract_info
                ie_result = ydl.sanitize_info(
                    ydl.extract_info(str(urlcanon.whatwg(ydl.url)))
                )
            metrics.brozzler_ydl_extract_successes.labels(ydl.is_youtube_host).inc(1)
            break
        except brozzler.ShutdownRequested:
            raise
        except Exception as e:
            if (
                hasattr(e, "exc_info")
                and e.exc_info[0] == yt_dlp.utils.UnsupportedError
            ):
                return None
            elif (
                hasattr(e, "exc_info")
                and e.exc_info[0] == urllib.error.HTTPError
                and hasattr(e.exc_info[1], "code")
                and e.exc_info[1].code == 420
            ):
                raise brozzler.ReachedLimit(e.exc_info[1])
            elif isinstance(e, yt_dlp.utils.DownloadError) and (
                "Redirect loop detected" in e.msg or "Too many redirects" in e.msg
            ):
                raise brozzler.VideoExtractorError(e.msg)
            else:
                # todo: other errors to handle separately?
                # OSError('Tunnel connection failed: 464 Host Not Allowed') (caused by ProxyError...)
                # and others...
                attempt += 1
                if attempt == max_attempts:
                    logger.warning(
                        "Failed after %s attempt(s)",
                        max_attempts,
                        attempts=max_attempts,
                        exc_info=True,
                    )
                    raise brozzler.VideoExtractorError(
                        "yt-dlp hit error extracting info for %s" % ydl.url
                    )
                else:
                    retry_wait = min(60, YTDLP_WAIT * (1.5 ** (attempt - 1)))
                    logger.info(
                        "Attempt %s failed. Retrying in %s seconds...",
                        attempt,
                        retry_wait,
                    )
                    time.sleep(retry_wait)
    else:
        raise brozzler.VideoExtractorError(
            "yt-dlp hit unknown error extracting info for %s" % ydl.url
        )

    logger.info("ytdlp completed successfully")

    _remember_videos(page, ydl.pushed_videos)
    if worker._using_warcprox(site):
        info_json = json.dumps(ie_result, sort_keys=True, indent=4)
        logger.info(
            "sending WARCPROX_WRITE_RECORD request to warcprox with yt-dlp json",
            url=ydl.url,
        )
        worker._warcprox_write_record(
            warcprox_address=worker._proxy_for(site),
            url="youtube-dl:%s" % str(urlcanon.semantic(ydl.url)),
            warc_type="metadata",
            content_type="application/vnd.youtube-dl_formats+json;charset=utf-8",
            payload=info_json.encode("utf-8"),
            extra_headers=site.extra_headers(page),
        )
    return ie_result


@metrics.brozzler_ytdlp_duration_seconds.time()
@metrics.brozzler_in_progress_ytdlps.track_inprogress()
def do_youtube_dl(worker, site, page, ytdlp_proxy_endpoints):
    """
    Runs yt-dlp configured for `worker` and `site` to download videos from
    `page`.

    Args:
        worker (brozzler.BrozzlerWorker): the calling brozzler worker
        site (brozzler.Site): the site we are brozzling
        page (brozzler.Page): the page we are brozzling

    Returns:
         `list` of `str`: outlink urls
    """
    with tempfile.TemporaryDirectory(
        prefix="brzl-ydl-", dir=worker._ytdlp_tmpdir
    ) as tempdir:
        logger.info("tempdir for yt-dlp", tempdir=tempdir)
        ydl = _build_youtube_dl(worker, tempdir, site, page, ytdlp_proxy_endpoints)
        ie_result = _try_youtube_dl(worker, ydl, site, page)
        outlinks = set()
        if ie_result and (
            ie_result.get("extractor") == "youtube:playlist"
            or ie_result.get("extractor") == "youtube:tab"
        ):
            # youtube watch pages as outlinks
            outlinks = {
                "https://www.youtube.com/watch?v=%s" % e["id"]
                for e in ie_result.get("entries_no_dl", [])
            }
        # any outlinks for other cases? soundcloud, maybe?
        return outlinks



"""
brozzler/robots.py - robots.txt support

Uses the reppy library version 0.3.4. Monkey-patches reppy to support substring
user-agent matching. We're sticking with 0.3.4 because later versions don't
support supplying a custom requests.Session.

See also https://github.com/seomoz/reppy/issues/37

Copyright (C) 2014-2016 Internet Archive

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json

import reppy
import reppy.cache
import reppy.parser
import requests
import structlog

import brozzler
from brozzler.ssl import CustomSSLContextHTTPAdapter, permissive_ssl_context

__all__ = ["is_permitted_by_robots"]

# monkey-patch reppy to do substring user-agent matching, see top of file
reppy.Utility.short_user_agent = lambda strng: strng


def _reppy_rules_getitem(self, agent):
    """
    Find the user-agent token matching the supplied full user-agent, using
    a case-insensitive substring search.
    """
    lc_agent = agent.lower()
    for s in self.agents:
        if s in lc_agent:
            return self.agents[s]
    return self.agents.get("*")


reppy.parser.Rules.__getitem__ = _reppy_rules_getitem


class _SessionRaiseOn420(requests.Session):
    timeout = 60

    def get(self, url, *args, **kwargs):
        res = super().get(url, timeout=self.timeout, *args, **kwargs)
        if res.status_code == 420 and "warcprox-meta" in res.headers:
            raise brozzler.ReachedLimit(
                warcprox_meta=json.loads(res.headers["warcprox-meta"]),
                http_payload=res.text,
            )
        else:
            return res


_robots_caches = {}  # {site_id:reppy.cache.RobotsCache}


def _robots_cache(site, proxy=None):
    if site.id not in _robots_caches:
        req_sesh = _SessionRaiseOn420()
        req_sesh.verify = False  # ignore cert errors

        ctx = permissive_ssl_context()  # allow unsafe SSL renegotiation
        ctx.check_hostname = False  # required when verify=False
        req_sesh.mount("https://", CustomSSLContextHTTPAdapter(ctx))

        if proxy:
            proxie = "http://%s" % proxy
            req_sesh.proxies = {"http": proxie, "https": proxie}
        if site.extra_headers():
            req_sesh.headers.update(site.extra_headers())
        if site.user_agent:
            req_sesh.headers["User-Agent"] = site.user_agent
        _robots_caches[site.id] = reppy.cache.RobotsCache(
            session=req_sesh, disallow_forbidden=False
        )

    return _robots_caches[site.id]


def is_permitted_by_robots(site, url, proxy=None):
    """
    Checks if `url` is permitted by robots.txt.

    Treats any kind of error fetching robots.txt as "allow all". See
    http://builds.archive.org/javadoc/heritrix-3.x-snapshot/org/archive/modules/net/CrawlServer.html#updateRobots(org.archive.modules.CrawlURI)
    for some background on that policy.

    Returns:
        bool: `True` if `site.ignore_robots` is set, or if `url` is permitted
            by robots.txt, `False` otherwise

    Raises:
        brozzler.ReachedLimit: if warcprox responded with 420 Reached Limit
        requests.exceptions.ProxyError: if the proxy is down
    """
    if site.ignore_robots:
        return True

    try:
        result = _robots_cache(site, proxy).allowed(url, site.user_agent or "brozzler")
        return result
    except Exception as e:
        if isinstance(e, reppy.exceptions.ServerError) and isinstance(
            e.args[0], brozzler.ReachedLimit
        ):
            raise e.args[0]
        elif hasattr(e, "args") and isinstance(
            e.args[0], requests.exceptions.ProxyError
        ):
            # reppy has wrapped an exception that we want to bubble up
            raise brozzler.ProxyError(e)
        else:
            structlog.get_logger(logger_name=__name__).warning(
                "returning true (permitted) after problem fetching robots.txt",
                url=url,
                raised_exception=e,
            )
            return True



import ssl

import requests
import urllib3


class CustomSSLContextHTTPAdapter(requests.adapters.HTTPAdapter):
    def __init__(self, ssl_context=None, **kwargs):
        self.ssl_context = ssl_context
        super().__init__(**kwargs)

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = urllib3.poolmanager.PoolManager(
            num_pools=connections,
            maxsize=maxsize,
            block=block,
            ssl_context=self.ssl_context,
        )


def permissive_ssl_context():
    ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
    ctx.options |= 0x4  # OP_LEGACY_SERVER_CONNECT

    return ctx
